# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from awscrt.s3 import S3RequestType

from awscli.compat import BytesIO
from awscli.customizations.s3.utils import relative_path
from awscli.testutils import cd, mock, skip_if_case_sensitive, skip_if_windows
from tests.functional.s3 import (
    BaseCRTTransferClientTest,
    BaseS3CLIRunnerTest,
    BaseS3TransferCommandTest,
)


class TestSyncCommand(BaseS3TransferCommandTest):
    prefix = 's3 sync '

    def test_website_redirect_ignore_paramfile(self):
        full_path = self.files.create_file('foo.txt', 'mycontent')
        cmdline = '%s %s s3://bucket/key.txt --website-redirect %s' % (
            self.prefix,
            self.files.rootdir,
            'http://someserver',
        )
        self.parsed_responses = [
            {"CommonPrefixes": [], "Contents": []},
            {'ETag': '"c8afdb36c52cf4727836669019e69222"'},
        ]
        self.run_cmd(cmdline, expected_rc=0)

        # The only operations we should have called are ListObjectsV2/PutObject.
        self.assertEqual(
            len(self.operations_called), 2, self.operations_called
        )
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        # Make sure that the specified web address is used as opposed to the
        # contents of the web address when uploading the object
        self.assertEqual(
            self.operations_called[1][1]['WebsiteRedirectLocation'],
            'http://someserver',
        )

    def test_no_recursive_option(self):
        cmdline = '. s3://mybucket --recursive'
        # Return code will be 252 for invalid parameter ``--recursive``
        self.run_cmd(cmdline, expected_rc=252)

    def test_sync_from_non_existant_directory(self):
        non_existant_directory = os.path.join(self.files.rootdir, 'fakedir')
        cmdline = '%s %s s3://bucket/' % (self.prefix, non_existant_directory)
        self.parsed_responses = [{"CommonPrefixes": [], "Contents": []}]
        _, stderr, _ = self.run_cmd(cmdline, expected_rc=255)
        self.assertIn('does not exist', stderr)

    def test_sync_to_non_existant_directory(self):
        key = 'foo.txt'
        non_existant_directory = os.path.join(self.files.rootdir, 'fakedir')
        cmdline = '%s s3://bucket/ %s' % (self.prefix, non_existant_directory)
        self.parsed_responses = [
            {
                "CommonPrefixes": [],
                "Contents": [
                    {
                        "Key": key,
                        "Size": 3,
                        "LastModified": "2014-01-09T20:45:49.000Z",
                        "ETag": '"c8afdb36c52cf4727836669019e69222-"',
                    }
                ],
            },
            {
                'ETag': '"c8afdb36c52cf4727836669019e69222-"',
                'Body': BytesIO(b'foo'),
            },
        ]
        self.run_cmd(cmdline, expected_rc=0)
        # Make sure the file now exists.
        self.assertTrue(
            os.path.exists(os.path.join(non_existant_directory, key))
        )

    def test_dryrun_sync(self):
        self.parsed_responses = [
            self.list_objects_response([]),
        ]
        full_path = self.files.create_file('file.txt', 'mycontent')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --dryrun'
        stdout, _, _ = self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                (
                    'ListObjectsV2',
                    {
                        'Bucket': 'bucket',
                        'Prefix': '',
                    },
                ),
            ]
        )
        self.assertIn(
            f'(dryrun) upload: {relative_path(full_path)} to '
            f's3://bucket/file.txt',
            stdout,
        )

    def test_glacier_sync_with_force_glacier(self):
        self.parsed_responses = [
            {
                'Contents': [
                    {
                        'Key': 'foo/bar.txt',
                        'ContentLength': '100',
                        'LastModified': '00:00:00Z',
                        'StorageClass': 'GLACIER',
                        'Size': 100,
                        'ETag': '"foo-1"',
                    },
                ],
                'CommonPrefixes': [],
            },
            {'ETag': '"foo-1"', 'Body': BytesIO(b'foo')},
        ]
        cmdline = '%s s3://bucket/foo %s --force-glacier-transfer' % (
            self.prefix,
            self.files.rootdir,
        )
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(
            len(self.operations_called), 2, self.operations_called
        )
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')

    def test_handles_glacier_incompatible_operations(self):
        self.parsed_responses = [
            {
                'Contents': [
                    {
                        'Key': 'foo',
                        'Size': 100,
                        'LastModified': '00:00:00Z',
                        'StorageClass': 'GLACIER',
                    },
                    {
                        'Key': 'bar',
                        'Size': 100,
                        'LastModified': '00:00:00Z',
                        'StorageClass': 'DEEP_ARCHIVE',
                    },
                ]
            }
        ]
        cmdline = '%s s3://bucket/ %s' % (self.prefix, self.files.rootdir)
        _, stderr, _ = self.run_cmd(cmdline, expected_rc=2)
        # There should not have been a download attempted because the
        # operation was skipped because it is glacier and glacier
        # deep archive incompatible.
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertIn('GLACIER', stderr)
        self.assertIn('s3://bucket/foo', stderr)
        self.assertIn('s3://bucket/bar', stderr)

    def test_turn_off_glacier_warnings(self):
        self.parsed_responses = [
            {
                'Contents': [
                    {
                        'Key': 'foo',
                        'Size': 100,
                        'LastModified': '00:00:00Z',
                        'StorageClass': 'GLACIER',
                    },
                    {
                        'Key': 'bar',
                        'Size': 100,
                        'LastModified': '00:00:00Z',
                        'StorageClass': 'DEEP_ARCHIVE',
                    },
                ]
            }
        ]
        cmdline = '%s s3://bucket/ %s --ignore-glacier-warnings' % (
            self.prefix,
            self.files.rootdir,
        )
        _, stderr, _ = self.run_cmd(cmdline, expected_rc=0)
        # There should not have been a download attempted because the
        # operation was skipped because it is glacier incompatible.
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual('', stderr)

    def test_warning_on_invalid_timestamp(self):
        full_path = self.files.create_file('foo.txt', 'mycontent')

        cmdline = '%s %s s3://bucket/key.txt' % (
            self.prefix,
            self.files.rootdir,
        )
        self.parsed_responses = [
            {"CommonPrefixes": [], "Contents": []},
            {'ETag': '"c8afdb36c52cf4727836669019e69222"'},
        ]
        # Patch get_file_stat to return a value indicating that an invalid
        # timestamp was loaded. It is impossible to set an invalid timestamp
        # on all OSes so it has to be patched.
        # TODO: find another method to test this behavior without patching.
        with mock.patch(
            'awscli.customizations.s3.filegenerator.get_file_stat',
            return_value=(None, None),
        ):
            self.run_cmd(cmdline, expected_rc=2)

        # We should still have put the object
        self.assertEqual(
            len(self.operations_called), 2, self.operations_called
        )
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')

    def test_sync_with_delete_on_downloads(self):
        full_path = self.files.create_file('foo.txt', 'mycontent')
        cmdline = '%s s3://bucket %s --delete' % (
            self.prefix,
            self.files.rootdir,
        )
        self.parsed_responses = [
            {"CommonPrefixes": [], "Contents": []},
            {'ETag': '"c8afdb36c52cf4727836669019e69222"'},
        ]
        self.run_cmd(cmdline, expected_rc=0)

        # The only operations we should have called are ListObjectsV2.
        self.assertEqual(
            len(self.operations_called), 1, self.operations_called
        )
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')

        self.assertFalse(os.path.exists(full_path))

    # When a file has been deleted after listing,
    # awscli.customizations.s3.utils.get_file_stat may raise either some kind
    # of OSError, or a ValueError, depending on the environment. In both cases,
    # the behaviour should be the same: skip the file and emit a warning.
    #
    # This test covers the case where a ValueError is emitted.
    def test_sync_skips_over_files_deleted_between_listing_and_transfer_valueerror(
        self,
    ):
        full_path = self.files.create_file('foo.txt', 'mycontent')
        cmdline = '%s %s s3://bucket/' % (self.prefix, self.files.rootdir)

        # FileGenerator.list_files should skip over files that cause an
        # IOError to be raised because they are missing when we try to
        # get their stats. This IOError is translated to a ValueError in
        # awscli.customizations.s3.utils.get_file_stat.
        def side_effect(_):
            os.remove(full_path)
            raise ValueError()

        with mock.patch(
            'awscli.customizations.s3.filegenerator.get_file_stat',
            side_effect=side_effect,
        ):
            self.run_cmd(cmdline, expected_rc=2)

        # We should not call PutObject because the file was deleted
        # before we could transfer it
        self.assertEqual(
            len(self.operations_called), 1, self.operations_called
        )
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')

    # This test covers the case where an OSError is emitted.
    def test_sync_skips_over_files_deleted_between_listing_and_transfer_oserror(
        self,
    ):
        full_path = self.files.create_file('foo.txt', 'mycontent')
        cmdline = '%s %s s3://bucket/' % (self.prefix, self.files.rootdir)

        # FileGenerator.list_files should skip over files that cause an
        # OSError to be raised because they are missing when we try to
        # get their stats.
        def side_effect(_):
            os.remove(full_path)
            raise OSError()

        with mock.patch(
            'awscli.customizations.s3.filegenerator.get_file_stat',
            side_effect=side_effect,
        ):
            self.run_cmd(cmdline, expected_rc=2)

        # We should not call PutObject because the file was deleted
        # before we could transfer it
        self.assertEqual(
            len(self.operations_called), 1, self.operations_called
        )
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')

    def test_request_payer(self):
        cmdline = '%s s3://sourcebucket/ s3://mybucket --request-payer' % (
            self.prefix
        )
        self.parsed_responses = [
            # Response for ListObjects on source bucket
            self.list_objects_response(['mykey']),
            # Response for ListObjects on destination bucket
            self.list_objects_response([]),
            self.copy_object_response(),
        ]
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request(
                    'sourcebucket', RequestPayer='requester'
                ),
                self.list_objects_request(
                    'mybucket', RequestPayer='requester'
                ),
                self.copy_object_request(
                    'sourcebucket',
                    'mykey',
                    'mybucket',
                    'mykey',
                    RequestPayer='requester',
                ),
            ]
        )

    def test_s3s3_sync_with_destination_sse_c(self):
        cmdline = (
            '%s s3://sourcebucket/ s3://mybucket '
            '--sse-c AES256 --sse-c-key destination-key' % self.prefix
        )
        self.parsed_responses = [
            self.list_objects_response(['mykey']),
            self.list_objects_response([]),
            self.copy_object_response(),
        ]
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request('sourcebucket'),
                self.list_objects_request('mybucket'),
                self.copy_object_request(
                    'sourcebucket',
                    'mykey',
                    'mybucket',
                    'mykey',
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='destination-key',
                ),
            ]
        )

    def test_s3s3_sync_with_destination_sse_c_multipart(self):
        mp_threshold = 8 * 1024 * 1024
        cmdline = (
            '%s s3://sourcebucket/ s3://mybucket '
            '--sse-c AES256 --sse-c-key destination-key' % self.prefix
        )
        self.parsed_responses = [
            self.list_objects_response(['mykey'], Size=mp_threshold),
            self.list_objects_response([]),
            # HeadObject from SetMetadataDirectivePropsSubscriber
            self.head_object_response(ContentLength=mp_threshold),
            self.get_object_tagging_response({}),
        ] + self.mp_copy_responses()
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request('sourcebucket'),
                self.list_objects_request('mybucket'),
                self.head_object_request(
                    'sourcebucket',
                    'mykey',
                    # no SSE-C params — source is unencrypted
                ),
                self.get_object_tagging_request('sourcebucket', 'mykey'),
                self.create_mpu_request(
                    'mybucket',
                    'mykey',
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='destination-key',
                ),
                self.upload_part_copy_request(
                    'sourcebucket',
                    'mykey',
                    'mybucket',
                    'mykey',
                    'upload_id',
                    PartNumber=mock.ANY,
                    CopySourceRange=mock.ANY,
                    CopySourceIfMatch=mock.ANY,
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='destination-key',
                ),
                self.complete_mpu_request(
                    'mybucket',
                    'mykey',
                    'upload_id',
                    num_parts=1,
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='destination-key',
                ),
            ]
        )

    def test_s3s3_sync_with_different_sse_c_keys(self):
        cmdline = (
            '%s s3://sourcebucket/ s3://mybucket '
            '--sse-c-copy-source AES256 --sse-c-copy-source-key source-key '
            '--sse-c AES256 --sse-c-key destination-key' % self.prefix
        )
        self.parsed_responses = [
            self.list_objects_response(['mykey']),
            self.list_objects_response([]),
            self.copy_object_response(),
        ]
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request('sourcebucket'),
                self.list_objects_request('mybucket'),
                self.copy_object_request(
                    'sourcebucket',
                    'mykey',
                    'mybucket',
                    'mykey',
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='destination-key',
                    CopySourceSSECustomerAlgorithm='AES256',
                    CopySourceSSECustomerKey='source-key',
                ),
            ]
        )

    def test_s3s3_sync_with_different_sse_c_keys_multipart(self):
        mp_threshold = 8 * 1024 * 1024
        cmdline = (
            '%s s3://sourcebucket/ s3://mybucket '
            '--sse-c-copy-source AES256 --sse-c-copy-source-key source-key '
            '--sse-c AES256 --sse-c-key destination-key' % self.prefix
        )
        self.parsed_responses = [
            self.list_objects_response(['mykey'], Size=mp_threshold),
            self.list_objects_response([]),
            # HeadObject from SetMetadataDirectivePropsSubscriber
            self.head_object_response(ContentLength=mp_threshold),
            self.get_object_tagging_response({}),
        ] + self.mp_copy_responses()
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request('sourcebucket'),
                self.list_objects_request('mybucket'),
                self.head_object_request(
                    'sourcebucket',
                    'mykey',
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='source-key',
                ),
                self.get_object_tagging_request('sourcebucket', 'mykey'),
                self.create_mpu_request(
                    'mybucket',
                    'mykey',
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='destination-key',
                ),
                self.upload_part_copy_request(
                    'sourcebucket',
                    'mykey',
                    'mybucket',
                    'mykey',
                    'upload_id',
                    PartNumber=mock.ANY,
                    CopySourceRange=mock.ANY,
                    CopySourceIfMatch=mock.ANY,
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='destination-key',
                    CopySourceSSECustomerAlgorithm='AES256',
                    CopySourceSSECustomerKey='source-key',
                ),
                self.complete_mpu_request(
                    'mybucket',
                    'mykey',
                    'upload_id',
                    num_parts=1,
                    SSECustomerAlgorithm='AES256',
                    SSECustomerKey='destination-key',
                ),
            ]
        )

    def test_request_payer_with_deletes(self):
        cmdline = '%s s3://sourcebucket/ s3://mybucket' % self.prefix
        cmdline += ' --request-payer'
        cmdline += ' --delete'
        self.parsed_responses = [
            # Response for ListObjects on source bucket
            self.list_objects_response([]),
            # Response for ListObjects on destination bucket
            self.list_objects_response(['key-to-delete']),
            self.delete_object_response(),
        ]
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request(
                    'sourcebucket', RequestPayer='requester'
                ),
                self.list_objects_request(
                    'mybucket', RequestPayer='requester'
                ),
                self.delete_object_request(
                    'mybucket', 'key-to-delete', RequestPayer='requester'
                ),
            ]
        )

    def test_with_accesspoint_arn(self):
        accesspoint_arn = (
            'arn:aws:s3:us-west-2:123456789012:accesspoint/endpoint'
        )
        cmdline = self.prefix
        cmdline += 's3://%s' % accesspoint_arn
        cmdline += ' %s' % self.files.rootdir
        self.parsed_responses = [
            self.list_objects_response(['mykey']),
            self.get_object_response(),
        ]
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request(accesspoint_arn),
                self.get_object_request(accesspoint_arn, 'mykey'),
            ]
        )

    def test_with_copy_props(self):
        cmdline = self.prefix
        cmdline += 's3://sourcebucket/ s3://bucket/'
        cmdline += ' --copy-props default'

        upload_id = 'upload_id'
        large_tag_set = {'tag-key': 'val' * 3000}
        metadata = {'tag-key': 'tag-value'}
        self.parsed_responses = [
            self.list_objects_response(keys=['key'], Size=8 * 1024**2),
            self.list_objects_response(keys=[]),
            self.head_object_response(
                Metadata=metadata,
            ),
            self.get_object_tagging_response(large_tag_set),
            self.create_mpu_response(upload_id),
            self.upload_part_copy_response(),
            self.complete_mpu_response(),
            self.put_object_tagging_response(),
        ]
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request('sourcebucket'),
                self.list_objects_request('bucket'),
                self.head_object_request('sourcebucket', 'key'),
                self.get_object_tagging_request('sourcebucket', 'key'),
                self.create_mpu_request('bucket', 'key', Metadata=metadata),
                self.upload_part_copy_request(
                    'sourcebucket',
                    'key',
                    'bucket',
                    'key',
                    upload_id,
                    CopySourceRange=mock.ANY,
                    PartNumber=1,
                    CopySourceIfMatch='"foo-1"',
                ),
                self.complete_mpu_request('bucket', 'key', upload_id, 1),
                self.put_object_tagging_request(
                    'bucket', 'key', large_tag_set
                ),
            ]
        )

    def test_upload_with_checksum_algorithm_sha1(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm SHA1'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'SHA1'
        )

    def test_copy_with_checksum_algorithm_update_sha1(self):
        cmdline = f'{self.prefix} s3://src-bucket/ s3://dest-bucket/ --checksum-algorithm SHA1'
        self.parsed_responses = [
            # Response for ListObjects on source bucket
            {
                'Contents': [
                    {
                        'Key': 'mykey',
                        'LastModified': '00:00:00Z',
                        'Size': 100,
                        'ChecksumAlgorithm': 'SHA1',
                        'ETag': 'myetag',
                    }
                ],
                'CommonPrefixes': [],
            },
            # Response for ListObjects on destination bucket
            self.list_objects_response([]),
            # Response for CopyObject
            {'ChecksumSHA1': 'sha1-checksum'},
        ]
        self.run_cmd(cmdline, expected_rc=0)
        self.assert_operations_called(
            [
                self.list_objects_request('src-bucket'),
                self.list_objects_request('dest-bucket'),
                (
                    'CopyObject',
                    {
                        'CopySource': {'Bucket': 'src-bucket', 'Key': 'mykey'},
                        'Bucket': 'dest-bucket',
                        'Key': 'mykey',
                        'ChecksumAlgorithm': 'SHA1',
                    },
                ),
            ]
        )

    def test_upload_with_checksum_algorithm_sha256(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm SHA256'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'SHA256'
        )

    def test_upload_with_checksum_algorithm_sha512(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm SHA512'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'SHA512'
        )

    def test_upload_with_checksum_algorithm_crc32(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm CRC32'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'CRC32'
        )

    def test_upload_with_checksum_algorithm_crc32c(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm CRC32C'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'CRC32C'
        )

    def test_upload_with_checksum_algorithm_crc64nvme(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm CRC64NVME'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'CRC64NVME'
        )

    def test_upload_with_checksum_algorithm_xxhash3(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm XXHASH3'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'XXHASH3'
        )

    def test_upload_with_checksum_algorithm_xxhash64(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm XXHASH64'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'XXHASH64'
        )

    def test_upload_with_checksum_algorithm_xxhash128(self):
        self.files.create_file('foo.txt', 'contents')
        cmdline = f'{self.prefix} {self.files.rootdir} s3://bucket/ --checksum-algorithm XXHASH128'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(
            self.operations_called[1][1]['ChecksumAlgorithm'], 'XXHASH128'
        )

    def test_download_with_checksum_mode_sha1(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumSHA1': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_download_with_checksum_mode_sha256(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumSHA256': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_download_with_checksum_mode_sha512(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumSHA512': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_download_with_checksum_mode_crc32(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumCRC32': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_download_with_checksum_mode_crc32c(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumCRC32C': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_download_with_checksum_mode_crc64nvme(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumCRC64NVME': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_download_with_checksum_mode_xxhash3(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumXXHASH3': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_download_with_checksum_mode_xxhash64(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumXXHASH64': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_download_with_checksum_mode_xxhash128(self):
        self.parsed_responses = [
            self.list_objects_response(['bucket']),
            # Mocked GetObject response with a checksum algorithm specified
            {
                'ETag': 'foo-1',
                'ChecksumXXHASH128': 'checksum',
                'Body': BytesIO(b'foo'),
            },
        ]
        cmdline = f'{self.prefix} s3://bucket/foo {self.files.rootdir} --checksum-mode ENABLED'
        self.run_cmd(cmdline, expected_rc=0)

        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        self.assertIn(
            ('ChecksumMode', 'ENABLED'), self.operations_called[1][1].items()
        )

    def test_sync_upload_with_no_overwrite_when_file_does_not_exist_at_destination(
        self,
    ):
        self.files.create_file("new_file.txt", "mycontent")
        self.parsed_responses = [
            self.list_objects_response(['file.txt']),
            {'ETag': '"c8afdb36c52cf4727836669019e69222"'},
        ]
        cmdline = (
            f'{self.prefix} {self.files.rootdir} s3://bucket --no-overwrite'
        )
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(len(self.operations_called), 2)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'PutObject')
        self.assertEqual(self.operations_called[1][1]['Key'], 'new_file.txt')

    def test_sync_upload_with_no_overwrite_when_file_exists_at_destination(
        self,
    ):
        self.files.create_file("new_file.txt", "mycontent")
        self.parsed_responses = [
            self.list_objects_response(['new_file.txt']),
        ]
        cmdline = (
            f'{self.prefix} {self.files.rootdir} s3://bucket --no-overwrite'
        )
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')

    def test_sync_download_with_no_overwrite_file_not_exists_at_destination(
        self,
    ):
        self.parsed_responses = [
            self.list_objects_response(['new_file.txt']),
            self.get_object_response(),
        ]
        cmdline = (
            f'{self.prefix} s3://bucket/ {self.files.rootdir} --no-overwrite'
        )
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(len(self.operations_called), 2)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'GetObject')
        local_file_path = os.path.join(self.files.rootdir, 'new_file.txt')
        self.assertTrue(os.path.exists(local_file_path))

    def test_sync_download_with_no_overwrite_file_exists_at_destination(self):
        self.files.create_file('file.txt', 'My content')
        self.parsed_responses = [
            self.list_objects_response(['file.txt']),
        ]
        cmdline = (
            f'{self.prefix} s3://bucket/ {self.files.rootdir} --no-overwrite'
        )
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')

    def test_sync_copy_with_no_overwrite_file_not_exists_at_destination(self):
        self.parsed_responses = [
            self.list_objects_response(['new_file.txt']),
            self.list_objects_response(['file1.txt']),
            self.copy_object_response(),
        ]
        cmdline = f'{self.prefix} s3://bucket/ s3://bucket2/ --no-overwrite'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(len(self.operations_called), 3)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[2][0].name, 'CopyObject')
        self.assertEqual(self.operations_called[2][1]['Key'], 'new_file.txt')

    def test_sync_copy_with_no_overwrite_file_exists_at_destination(self):
        self.parsed_responses = [
            self.list_objects_response(['new_file.txt']),
            self.list_objects_response(['new_file.txt', 'file1.txt']),
        ]
        cmdline = f'{self.prefix} s3://bucket/ s3://bucket2/ --no-overwrite'
        self.run_cmd(cmdline, expected_rc=0)
        self.assertEqual(len(self.operations_called), 2)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')
        self.assertEqual(self.operations_called[1][0].name, 'ListObjectsV2')


class TestSyncSourceRegion(BaseS3CLIRunnerTest):
    def test_respects_source_region(self):
        source_region = 'af-south-1'
        cmdline = [
            's3',
            'sync',
            's3://sourcebucket/',
            's3://bucket/',
            '--region',
            self.region,
            '--source-region',
            source_region,
        ]
        self.add_botocore_list_objects_response(['key'])
        self.add_botocore_list_objects_response([])
        self.add_botocore_copy_object_response()
        result = self.run_command(cmdline)
        self.assert_no_remaining_botocore_responses()
        self.assert_operations_to_endpoints(
            cli_runner_result=result,
            expected_operations_to_endpoints=[
                (
                    'ListObjectsV2',
                    self.get_virtual_s3_host('sourcebucket', source_region),
                ),
                (
                    'ListObjectsV2',
                    self.get_virtual_s3_host('bucket', self.region),
                ),
                (
                    'CopyObject',
                    self.get_virtual_s3_host('bucket', self.region),
                ),
            ],
        )


class TestSyncWithCRTClient(BaseCRTTransferClientTest):
    def test_upload_sync_using_crt_client(self):
        filename = self.files.create_file('myfile', 'mycontent')
        cmdline = [
            's3',
            'sync',
            self.files.rootdir,
            's3://bucket/',
        ]
        self.add_botocore_list_objects_response([])
        self.run_command(cmdline)
        crt_requests = self.get_crt_make_request_calls()
        self.assertEqual(len(crt_requests), 1)
        self.assert_crt_make_request_call(
            crt_requests[0],
            expected_type=S3RequestType.PUT_OBJECT,
            expected_host=self.get_virtual_s3_host('bucket'),
            expected_path='/myfile',
            expected_send_filepath=filename,
        )

    def test_download_sync_using_crt_client(self):
        cmdline = [
            's3',
            'sync',
            's3://bucket/',
            self.files.rootdir,
        ]
        self.add_botocore_list_objects_response(['key'])
        self.run_command(cmdline)
        crt_requests = self.get_crt_make_request_calls()
        self.assertEqual(len(crt_requests), 1)
        self.assert_crt_make_request_call(
            crt_requests[0],
            expected_type=S3RequestType.GET_OBJECT,
            expected_host=self.get_virtual_s3_host('bucket'),
            expected_path='/key',
            expected_recv_startswith=os.path.join(self.files.rootdir, 'key'),
        )

    def test_upload_sync_with_delete_using_crt_client(self):
        filename = self.files.create_file('a-file', 'mycontent')
        cmdline = [
            's3',
            'sync',
            self.files.rootdir,
            's3://bucket/',
            '--delete',
        ]
        self.add_botocore_list_objects_response(['delete-this'])
        self.run_command(cmdline)
        crt_requests = self.get_crt_make_request_calls()
        self.assertEqual(len(crt_requests), 2)
        self.assert_crt_make_request_call(
            crt_requests[0],
            expected_type=S3RequestType.PUT_OBJECT,
            expected_host=self.get_virtual_s3_host('bucket'),
            expected_path='/a-file',
            expected_send_filepath=filename,
        )
        self.assert_crt_make_request_call(
            crt_requests[1],
            expected_type=S3RequestType.DEFAULT,
            expected_host=self.get_virtual_s3_host('bucket'),
            expected_path='/delete-this',
            expected_http_method='DELETE',
        )

    def test_download_sync_with_delete_using_crt_client(self):
        self.files.create_file('delete-this', 'content')
        cmdline = [
            's3',
            'sync',
            's3://bucket/',
            self.files.rootdir,
            '--delete',
        ]
        self.add_botocore_list_objects_response(['key'])
        self.run_command(cmdline)
        crt_requests = self.get_crt_make_request_calls()
        self.assertEqual(len(crt_requests), 1)
        self.assert_crt_make_request_call(
            crt_requests[0],
            expected_type=S3RequestType.GET_OBJECT,
            expected_host=self.get_virtual_s3_host('bucket'),
            expected_path='/key',
            expected_recv_startswith=os.path.join(self.files.rootdir, 'key'),
        )
        self.assertFalse(os.path.exists('delete-this'))

    def test_does_not_use_crt_client_for_copy_syncs(self):
        cmdline = ['s3', 'sync', 's3://bucket/', 's3://otherbucket/']
        self.add_botocore_list_objects_response(['key'])
        self.add_botocore_list_objects_response([])
        self.add_botocore_copy_object_response()
        self.run_command(cmdline)
        self.assertEqual(self.get_crt_make_request_calls(), [])
        self.assert_no_remaining_botocore_responses()


class TestSyncCommandWithS3Express(BaseS3TransferCommandTest):
    prefix = 's3 sync '

    def test_incompatible_with_sync_upload(self):
        cmdline = '%s %s s3://testdirectorybucket--usw2-az1--x-s3/' % (
            self.prefix,
            self.files.rootdir,
        )
        stderr = self.run_cmd(cmdline, expected_rc=252)[1]
        self.assertIn(
            'Cannot use sync command with a directory bucket.', stderr
        )

    def test_incompatible_with_sync_download(self):
        cmdline = '%s s3://testdirectorybucket--usw2-az1--x-s3/ %s' % (
            self.prefix,
            self.files.rootdir,
        )
        stderr = self.run_cmd(cmdline, expected_rc=252)[1]
        self.assertIn(
            'Cannot use sync command with a directory bucket.', stderr
        )

    def test_incompatible_with_sync_copy(self):
        cmdline = (
            '%s s3://bucket/ s3://testdirectorybucket--usw2-az1--x-s3/'
            % self.prefix
        )
        stderr = self.run_cmd(cmdline, expected_rc=252)[1]
        self.assertIn(
            'Cannot use sync command with a directory bucket.', stderr
        )

    def test_incompatible_with_sync_with_delete(self):
        cmdline = (
            '%s s3://bucket/ s3://testdirectorybucket--usw2-az1--x-s3/ --delete'
            % self.prefix
        )
        stderr = self.run_cmd(cmdline, expected_rc=252)[1]
        self.assertIn(
            'Cannot use sync command with a directory bucket.', stderr
        )

    def test_compatible_with_sync_with_local_directory_like_directory_bucket(
        self,
    ):
        self.parsed_responses = [{'Contents': []}]
        cmdline = (
            '%s s3://bucket/ testdirectorybucket--usw2-az1--x-s3/'
            % self.prefix
        )
        with cd(self.files.rootdir):
            _, stderr, _ = self.run_cmd(cmdline)
        # Just asserting that command validated and made an API call
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(self.operations_called[0][0].name, 'ListObjectsV2')


class TestSyncCaseConflict(BaseS3TransferCommandTest):
    prefix = 's3 sync '

    def setUp(self):
        super().setUp()
        self.lower_key = 'a.txt'
        self.upper_key = 'A.txt'

    @skip_if_case_sensitive()
    def test_error_with_existing_file(self):
        self.files.create_file(self.lower_key, 'mycontent')
        cmd = (
            f"{self.prefix} s3://bucket {self.files.rootdir} "
            "--case-conflict error"
        )
        self.parsed_responses = [self.list_objects_response([self.upper_key])]
        _, stderr, _ = self.run_cmd(cmd, expected_rc=1)
        assert f"Failed to download bucket/{self.upper_key}" in stderr

    def test_error_with_case_conflicts_in_s3(self):
        cmd = (
            f"{self.prefix} s3://bucket {self.files.rootdir} "
            "--case-conflict error"
        )
        self.parsed_responses = [
            self.list_objects_response([self.upper_key, self.lower_key])
        ]
        _, stderr, _ = self.run_cmd(cmd, expected_rc=1)
        assert f"Failed to download bucket/{self.lower_key}" in stderr

    @skip_if_case_sensitive()
    def test_warn_with_existing_file(self):
        self.files.create_file(self.lower_key, 'mycontent')
        cmd = (
            f"{self.prefix} s3://bucket {self.files.rootdir} "
            "--case-conflict warn"
        )
        self.parsed_responses = [
            self.list_objects_response([self.upper_key]),
            self.get_object_response(),
        ]
        _, stderr, _ = self.run_cmd(cmd, expected_rc=0)
        assert f"warning: Downloading bucket/{self.upper_key}" in stderr

    @skip_if_windows("Can't rename to same file")
    def test_warn_with_case_conflicts_in_s3(self):
        cmd = (
            f"{self.prefix} s3://bucket {self.files.rootdir} "
            "--case-conflict warn"
        )
        self.parsed_responses = [
            self.list_objects_response([self.upper_key, self.lower_key]),
            self.get_object_response(),
            self.get_object_response(),
        ]
        _, stderr, _ = self.run_cmd(cmd, expected_rc=0)
        assert f"warning: Downloading bucket/{self.lower_key}" in stderr

    @skip_if_case_sensitive()
    def test_skip_with_existing_file(self):
        self.files.create_file(self.lower_key, 'mycontent')
        cmd = (
            f"{self.prefix} s3://bucket {self.files.rootdir} "
            "--case-conflict skip"
        )
        self.parsed_responses = [self.list_objects_response([self.upper_key])]
        _, stderr, _ = self.run_cmd(cmd, expected_rc=0)
        assert f"warning: Skipping bucket/{self.upper_key}" in stderr

    def test_skip_with_case_conflicts_in_s3(self):
        cmd = (
            f"{self.prefix} s3://bucket {self.files.rootdir} "
            "--case-conflict skip"
        )
        self.parsed_responses = [
            self.list_objects_response([self.upper_key, self.lower_key]),
            self.get_object_response(),
        ]
        _, stderr, _ = self.run_cmd(cmd, expected_rc=0)
        assert f"warning: Skipping bucket/{self.lower_key}" in stderr

    def test_ignore_with_existing_file(self):
        self.files.create_file(self.lower_key, 'mycontent')
        cmd = (
            f"{self.prefix} s3://bucket {self.files.rootdir} "
            "--case-conflict ignore"
        )
        self.parsed_responses = [
            self.list_objects_response([self.upper_key]),
            self.get_object_response(),
        ]
        self.run_cmd(cmd, expected_rc=0)

    @skip_if_windows("Can't rename to same file")
    def test_ignore_with_case_conflicts_in_s3(self):
        cmd = (
            f"{self.prefix} s3://bucket {self.files.rootdir} "
            "--case-conflict ignore"
        )
        self.parsed_responses = [
            self.list_objects_response([self.upper_key, self.lower_key]),
            self.get_object_response(),
            self.get_object_response(),
        ]
        self.run_cmd(cmd, expected_rc=0)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.layout import (
    ConditionalContainer,
    FloatContainer,
    HSplit,
    Window,
)
from prompt_toolkit.widgets import Dialog

from awscli.autoprompt import widgets
from awscli.testutils import mock, unittest


class TestHelpPanelWidget(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.HelpPanelWidget()
        self.assertIsInstance(widget.container, ConditionalContainer)


class TestToolbarWidget(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.ToolbarWidget()
        self.assertIsInstance(widget.container, HSplit)


class TestHistorySignToolbarView(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.HistorySignToolbarView()
        self.assertIsInstance(widget.content, Window)

    def test_buffer_has_correct_wording(self):
        widget = widgets.HistorySignToolbarView()
        buffer = widget.content.content.buffer
        self.assertEqual(buffer.document.text, widget.help_text)


class TestInputToolbarView(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.InputToolbarView()
        self.assertIsInstance(widget.content, Window)

    def test_buffer_has_correct_wording(self):
        widget = widgets.InputToolbarView()
        buffer = widget.content.content.buffer
        self.assertEqual(buffer.document.text, widget.help_text)


class TestDocToolbarView(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.DocToolbarView()
        self.assertIsInstance(widget.content, Window)

    def test_buffer_has_correct_wording(self):
        widget = widgets.DocToolbarView()
        buffer = widget.content.content.buffer
        self.assertEqual(buffer.document.text, widget.help_text)


class TestDebugToolbarView(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.DebugToolbarView()
        self.assertIsInstance(widget.content, Window)

    def test_buffer_has_correct_wording(self):
        widget = widgets.DebugToolbarView()
        buffer = widget.content.content.buffer
        self.assertEqual(buffer.document.text, widget.help_text)


class TestDocHelpView(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.DocHelpView()
        self.assertIsInstance(widget.content, HSplit)

    def test_frame_has_correct_title(self):
        widget = widgets.DocHelpView()
        title = widget.create_window(Buffer()).title
        self.assertEqual(title, widget.TITLE)

    def test_buffer_has_correct_wording(self):
        widget = widgets.DocHelpView()
        buffer = widget.create_buffer()
        self.assertEqual(buffer.document.text, widget.help_text)


class TestInputHelpView(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.InputHelpView()
        self.assertIsInstance(widget.content, HSplit)

    def test_frame_has_correct_title(self):
        widget = widgets.InputHelpView()
        title = widget.create_window(Buffer()).title
        self.assertEqual(title, widget.TITLE)

    def test_buffer_has_correct_wording(self):
        widget = widgets.InputHelpView()
        buffer = widget.create_buffer()
        self.assertEqual(buffer.document.text, widget.help_text)


class TestDebugPanelWidget(unittest.TestCase):
    def test_can_create_container(self):
        widget = widgets.DebugPanelWidget()
        self.assertIsInstance(widget.container, ConditionalContainer)

    def test_can_create_save_dialog(self):
        widget = widgets.DebugPanelWidget()
        save_dialog = widget.create_save_file_dialog()
        self.assertIsInstance(save_dialog, Dialog)

    def test_can_create_float_container_with_correct_bindings(self):
        widget = widgets.DebugPanelWidget()
        self.assertIsInstance(widget.float_container, FloatContainer)
        key_bindings = widget.float_container.key_bindings.get_bindings_starting_with_keys(
            ''
        )
        self.assertEqual(len(key_bindings), 1)
        self.assertEqual(key_bindings[0].keys, ('c-s',))


class TestFormatTextProcessor(unittest.TestCase):
    def setUp(self):
        self.format_text_processor = widgets.FormatTextProcessor()

    def test_can_apply_transformation(self):
        text_input = mock.Mock()
        text_input.fragments = [('', '<style fg="black">[tab]</style>text1')]
        transformed_text = self.format_text_processor.apply_transformation(
            text_input
        )
        actual_fragments = transformed_text.fragments
        expected_fragments = FormattedText(
            [('fg:black', '[tab]'), ('', 'text1')]
        )
        self.assertEqual(actual_fragments, expected_fragments)

    def test_can_apply_transformation_on_multiple_tags(self):
        text_input = mock.Mock()
        text_input.fragments = [
            (
                '',
                (
                    '<style fg="black">[tab]</style>text1'
                    '<style fg="red">[up]</style>text2'
                    '<style fg="blue">[down]</style>text3'
                ),
            )
        ]
        transformed_text = self.format_text_processor.apply_transformation(
            text_input
        )
        actual_fragments = transformed_text.fragments
        expected_fragments = FormattedText(
            [
                ('fg:black', '[tab]'),
                ('', 'text1'),
                ('fg:red', '[up]'),
                ('', 'text2'),
                ('fg:blue', '[down]'),
                ('', 'text3'),
            ]
        )
        self.assertEqual(actual_fragments, expected_fragments)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime

from botocore.compat import urlsplit

from awscli.clidriver import AWSCLIEntryPoint
from awscli.testutils import (
    BaseAWSCommandParamsTest,
    create_clidriver,
    mock,
    temporary_file,
)

# Values used to fix time.time() and datetime.datetime.utcnow()
# so we know the exact values of the signatures generated.
FROZEN_TIMESTAMP = 1471305652
DEFAULT_EXPIRES = 3600
FROZEN_TIME = mock.Mock(return_value=FROZEN_TIMESTAMP)
FROZEN_DATETIME = mock.Mock(
    return_value=datetime.datetime(2016, 8, 18, 14, 33, 3, 0)
)


class TestPresignCommand(BaseAWSCommandParamsTest):
    prefix = 's3 presign '

    def enable_addressing_mode_in_config(self, fileobj, mode):
        fileobj.write(
            "[default]\n" "s3 =\n" "    addressing_style = %s\n" % mode
        )
        fileobj.flush()
        self.environ['AWS_CONFIG_FILE'] = fileobj.name
        self.driver = create_clidriver()
        self.entry_point = AWSCLIEntryPoint(self.driver)

    def enable_sigv4_from_config_file(self, fileobj):
        fileobj.write("[default]\n" "s3 =\n" "    signature_version = s3v4\n")
        fileobj.flush()
        self.environ['AWS_CONFIG_FILE'] = fileobj.name
        self.driver = create_clidriver()
        self.entry_point = AWSCLIEntryPoint(self.driver)

    def assert_presigned_url_matches(self, actual_url, expected_match):
        """Verify generated presigned URL matches expected dict.

        This method compares an actual URL against a dict of expected
        values.  The reason that the "expected_match" is a dict instead
        of the expected presigned URL is because the query params
        are unordered so we can't guarantee an expected query param
        ordering.

        """
        parts = urlsplit(actual_url)
        self.assertEqual(parts.netloc, expected_match['hostname'])
        self.assertEqual(parts.path, expected_match['path'])
        query_params = self.parse_query_string(parts.query)
        self.assertEqual(query_params, expected_match['query_params'])

    def parse_query_string(self, query_string):
        pairs = []
        for part in query_string.split('&'):
            pairs.append(part.split('=', 1))
        return dict(pairs)

    def get_presigned_url_for_cmd(self, cmdline):
        with mock.patch('time.time', FROZEN_TIME):
            with mock.patch('datetime.datetime') as d:
                d.utcnow = FROZEN_DATETIME
                stdout = self.assert_params_for_cmd(cmdline, None)[0].strip()
                return stdout

    def test_generates_a_url(self):
        stdout = self.get_presigned_url_for_cmd(
            self.prefix + 's3://bucket/key'
        )

        self.assert_presigned_url_matches(
            stdout,
            {
                'hostname': 'bucket.s3.us-east-1.amazonaws.com',
                'path': '/key',
                'query_params': {
                    'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
                    'X-Amz-Credential': (
                        'access_key%2F20160818%2Fus-east-1'
                        '%2Fs3%2Faws4_request'
                    ),
                    'X-Amz-Date': '20160818T143303Z',
                    'X-Amz-Expires': '3600',
                    'X-Amz-Signature': (
                        '1297528058f2c8b89cfa52c6a47d6c54890700a1da2470'
                        '2b06d53e774c0acc95'
                    ),
                    'X-Amz-SignedHeaders': 'host',
                },
            },
        )

    def test_handles_non_dns_compatible_buckets(self):
        stdout = self.get_presigned_url_for_cmd(
            self.prefix + 's3://bucket.dots/key'
        )

        self.assert_presigned_url_matches(
            stdout,
            {
                'hostname': 's3.us-east-1.amazonaws.com',
                'path': '/bucket.dots/key',
                'query_params': {
                    'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
                    'X-Amz-Credential': (
                        'access_key%2F20160818%2Fus-east-1'
                        '%2Fs3%2Faws4_request'
                    ),
                    'X-Amz-Date': '20160818T143303Z',
                    'X-Amz-Expires': '3600',
                    'X-Amz-Signature': (
                        '5a032639cabfe3db0b4b87ba3b12c29f5e42fe74cbba8'
                        'a0eb69bfb30c6e2d277'
                    ),
                    'X-Amz-SignedHeaders': 'host',
                },
            },
        )

    def test_handles_expires_in(self):
        expires_in = 1000
        stdout = self.get_presigned_url_for_cmd(
            self.prefix + 's3://bucket/key --expires-in %s' % expires_in
        )

        self.assert_presigned_url_matches(
            stdout,
            {
                'hostname': 'bucket.s3.us-east-1.amazonaws.com',
                'path': '/key',
                'query_params': {
                    'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
                    'X-Amz-Credential': (
                        'access_key%2F20160818%2Fus-east-1'
                        '%2Fs3%2Faws4_request'
                    ),
                    'X-Amz-Date': '20160818T143303Z',
                    'X-Amz-Expires': f'{expires_in}',
                    'X-Amz-Signature': (
                        '865fb61b021c3bf406c40d41353f584835fff1f158cf1b'
                        '3e6ec06260ecbb8937'
                    ),
                    'X-Amz-SignedHeaders': 'host',
                },
            },
        )

    def test_handles_sigv4(self):
        with temporary_file('w') as f:
            self.enable_sigv4_from_config_file(f)
            stdout = self.get_presigned_url_for_cmd(
                self.prefix + 's3://bucket/key'
            )

        expected = {
            'hostname': 'bucket.s3.us-east-1.amazonaws.com',
            'path': '/key',
            'query_params': {
                'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
                'X-Amz-Credential': (
                    'access_key%2F20160818%2Fus-east-1' '%2Fs3%2Faws4_request'
                ),
                'X-Amz-Date': '20160818T143303Z',
                'X-Amz-Expires': '3600',
                'X-Amz-Signature': (
                    '1297528058f2c8b89cfa52c6a47d6c548907'
                    '00a1da24702b06d53e774c0acc95'
                ),
                'X-Amz-SignedHeaders': 'host',
            },
        }
        self.assert_presigned_url_matches(stdout, expected)

    def test_s3_prefix_not_needed(self):
        # Consistent with the 'ls' command.
        stdout = self.get_presigned_url_for_cmd(self.prefix + 'bucket/key')

        self.assert_presigned_url_matches(
            stdout,
            {
                'hostname': 'bucket.s3.us-east-1.amazonaws.com',
                'path': '/key',
                'query_params': {
                    'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
                    'X-Amz-Credential': (
                        'access_key%2F20160818%2Fus-east-1'
                        '%2Fs3%2Faws4_request'
                    ),
                    'X-Amz-Date': '20160818T143303Z',
                    'X-Amz-Expires': '3600',
                    'X-Amz-Signature': (
                        '1297528058f2c8b89cfa52c6a47d6c54890700a1da2470'
                        '2b06d53e774c0acc95'
                    ),
                    'X-Amz-SignedHeaders': 'host',
                },
            },
        )

    def test_can_support_addressing_mode_config(self):
        with temporary_file('w') as f:
            self.enable_addressing_mode_in_config(f, 'path')
            stdout = self.get_presigned_url_for_cmd(
                self.prefix + 's3://bucket/key'
            )
        self.assert_presigned_url_matches(
            stdout,
            {
                'hostname': 's3.us-east-1.amazonaws.com',
                'path': '/bucket/key',
                'query_params': {
                    'X-Amz-Algorithm': 'AWS4-HMAC-SHA256',
                    'X-Amz-Credential': (
                        'access_key%2F20160818%2Fus-east-1'
                        '%2Fs3%2Faws4_request'
                    ),
                    'X-Amz-Date': '20160818T143303Z',
                    'X-Amz-Expires': '3600',
                    'X-Amz-Signature': (
                        'c6dab3560db76aded03e6268338ddb0a6dec00ebc82d6e'
                        '7abdc305529fcaba74'
                    ),
                    'X-Amz-SignedHeaders': 'host',
                },
            },
        )



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import copy
import json

from awscli.customizations.emrcontainers.base36 import Base36
from awscli.customizations.emrcontainers.constants import (
    TRUST_POLICY_STATEMENT_ALREADY_EXISTS,
    TRUST_POLICY_STATEMENT_FORMAT,
    TRUST_POLICY_UPDATE_SUCCESSFUL,
)
from awscli.testutils import BaseAWSCommandParamsTest, mock, unittest


def json_matches(first, second):
    return json.dumps(first, sort_keys=True) == json.dumps(
        second, sort_keys=True
    )


class TestUpdateAssumeRolePolicy(BaseAWSCommandParamsTest):
    cluster_name = 'test-cluster'
    namespace = 'test'
    role_name = 'myrole'
    account_id = '123456789012'
    oidc_provider = 'oidc-provider/id/test'
    aws_partition = 'aws'
    aws_cn_partition = 'aws-cn'

    base36_encoded_role_name = Base36().encode(role_name)
    expected_statement = TRUST_POLICY_STATEMENT_FORMAT % {
        "AWS_ACCOUNT_ID": account_id,
        "OIDC_PROVIDER": oidc_provider,
        "NAMESPACE": namespace,
        "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
        "AWS_PARTITION": aws_partition,
    }

    expected_statement_cn = TRUST_POLICY_STATEMENT_FORMAT % {
        "AWS_ACCOUNT_ID": account_id,
        "OIDC_PROVIDER": oidc_provider,
        "NAMESPACE": namespace,
        "BASE36_ENCODED_ROLE_NAME": base36_encoded_role_name,
        "AWS_PARTITION": aws_cn_partition,
    }

    def setUp(self):
        super(TestUpdateAssumeRolePolicy, self).setUp()

        self.command = (
            'emr-containers update-role-trust-policy --cluster-name=%s '
            '--namespace=%s --role-name=%s'
            % (self.cluster_name, self.namespace, self.role_name)
        )
        self.policy_document = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"AWS": "arn:aws:iam::123456789012:root"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }

        self.expected_policy_document = copy.deepcopy(self.policy_document)
        self.expected_policy_document.get("Statement").append(
            json.loads(self.expected_statement)
        )

        self.expected_policy_document_cn = copy.deepcopy(self.policy_document)
        self.expected_policy_document_cn.get("Statement").append(
            json.loads(self.expected_statement_cn)
        )

    # Assert the call to update trust policy of the role
    def assert_trust_policy_updated(
        self, cmd_output, expected_policy_document=None
    ):
        if expected_policy_document is None:
            expected_policy_document = self.expected_policy_document

        self.assertTrue(
            TRUST_POLICY_UPDATE_SUCCESSFUL % self.role_name in cmd_output
        )

        # Check if UpdateAssumeRolePolicy was invoked
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(
            self.operations_called[0][0].name, 'UpdateAssumeRolePolicy'
        )
        self.assertEqual(
            self.operations_called[0][1]['RoleName'], self.role_name
        )

        self.assertTrue(
            json_matches(
                json.loads(self.operations_called[0][1]['PolicyDocument']),
                expected_policy_document,
            )
        )

    # Use case: Expected trust policy does not exist
    # Expected results: Operation is performed by client
    # to update the trust policy in expected format
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'iam.IAM.get_assume_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_oidc_issuer_id'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_account_id'
    )
    def test_trust_policy_does_not_exist(
        self,
        get_account_id_patch,
        get_oidc_issuer_id_patch,
        get_assume_role_policy_patch,
    ):
        get_assume_role_policy_patch.return_value = self.policy_document
        get_oidc_issuer_id_patch.return_value = self.oidc_provider
        get_account_id_patch.return_value = self.account_id

        output = self.run_cmd(self.command, expected_rc=0)
        self.assert_trust_policy_updated(output[0])

    # Use case: Expected trust policy exists but the condition section
    # has an additional condition
    # Expected results: Operation is performed by client to update
    # the trust policy in expected format
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'iam.IAM.get_assume_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_oidc_issuer_id'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_account_id'
    )
    def test_trust_policy_exists_with_more_keys(
        self,
        get_account_id_patch,
        get_oidc_issuer_id_patch,
        get_assume_role_policy_patch,
    ):
        statement_with_additional_condition_key = json.loads(
            self.expected_statement
        )
        statement_with_additional_condition_key.get("Condition").get(
            "StringLike"
        )["test:key"] = "value"
        self.policy_document.get("Statement").append(
            statement_with_additional_condition_key
        )

        self.expected_policy_document = copy.deepcopy(self.policy_document)
        self.expected_policy_document.get("Statement").append(
            json.loads(self.expected_statement)
        )

        get_assume_role_policy_patch.return_value = self.policy_document
        get_oidc_issuer_id_patch.return_value = self.oidc_provider
        get_account_id_patch.return_value = self.account_id

        output = self.run_cmd(self.command, expected_rc=0)
        self.assert_trust_policy_updated(output[0])

    # Use case: Initial trust policy document does not have Statements section
    # Expected results: Operation is performed by client to update
    # the trust policy in expected format
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'iam.IAM.get_assume_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_oidc_issuer_id'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_account_id'
    )
    def test_policy_document_has_missing_key(
        self,
        get_account_id_patch,
        get_oidc_issuer_id_patch,
        get_assume_role_policy_patch,
    ):
        del self.policy_document["Statement"]

        self.expected_policy_document = copy.deepcopy(self.policy_document)
        self.expected_policy_document["Statement"] = [
            json.loads(self.expected_statement)
        ]

        get_assume_role_policy_patch.return_value = self.policy_document
        get_oidc_issuer_id_patch.return_value = self.oidc_provider
        get_account_id_patch.return_value = self.account_id

        output = self.run_cmd(self.command, expected_rc=0)
        self.assert_trust_policy_updated(output[0])

    # Use case: Initial trust policy document has empty Statements section
    # Expected results: Operation is performed by client to update
    # the trust policy in expected format
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'iam.IAM.get_assume_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_oidc_issuer_id'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_account_id'
    )
    def test_policy_document_has_empty_statements(
        self,
        get_account_id_patch,
        get_oidc_issuer_id_patch,
        get_assume_role_policy_patch,
    ):
        del self.policy_document.get("Statement")[:]

        self.expected_policy_document = copy.deepcopy(self.policy_document)
        self.expected_policy_document.get("Statement").append(
            json.loads(self.expected_statement)
        )

        get_assume_role_policy_patch.return_value = self.policy_document
        get_oidc_issuer_id_patch.return_value = self.oidc_provider
        get_account_id_patch.return_value = self.account_id

        output = self.run_cmd(self.command, expected_rc=0)
        self.assert_trust_policy_updated(output[0])

    # Use case: Expected trust policy does not exist and user performs a dry run
    # Expected results: No operation is performed by client
    # The command should print the expected policy document to stdout
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'iam.IAM.get_assume_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_oidc_issuer_id'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_account_id'
    )
    def test_trust_policy_does_not_exist_dry_run(
        self,
        get_account_id_patch,
        get_oidc_issuer_id_patch,
        get_assume_role_policy_patch,
    ):
        get_assume_role_policy_patch.return_value = self.policy_document
        get_oidc_issuer_id_patch.return_value = self.oidc_provider
        get_account_id_patch.return_value = self.account_id

        output = self.run_cmd(self.command + " --dry-run", expected_rc=0)
        self.assertEqual(len(self.operations_called), 0)
        self.assertTrue(
            json_matches(json.loads(output[0]), self.expected_policy_document)
        )

    # Use case: Expected trust policy already exists
    # Expected results: No operation is performed by client
    # The command should print that the trust policy statement already exists
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'iam.IAM.get_assume_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_oidc_issuer_id'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_account_id'
    )
    def test_trust_policy_exists(
        self,
        get_account_id_patch,
        get_oidc_issuer_id_patch,
        get_assume_role_policy_patch,
    ):
        self.policy_document = self.expected_policy_document

        get_assume_role_policy_patch.return_value = self.policy_document
        get_oidc_issuer_id_patch.return_value = self.oidc_provider
        get_account_id_patch.return_value = self.account_id

        output = self.run_cmd(self.command, expected_rc=0)
        self.assertEqual(len(self.operations_called), 0)
        self.assertTrue(
            TRUST_POLICY_STATEMENT_ALREADY_EXISTS % self.role_name in output[0]
        )

    # Use case: Expected trust policy does not exist in cn-north-1
    # Expected results: Operation is performed by client in cn-north-1
    # to update the trust policy in expected format
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'iam.IAM.get_assume_role_policy'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_oidc_issuer_id'
    )
    @mock.patch(
        'awscli.customizations.emrcontainers.' 'eks.EKS.get_account_id'
    )
    def test_trust_policy_does_not_exist_in_cn(
        self,
        get_account_id_patch,
        get_oidc_issuer_id_patch,
        get_assume_role_policy_patch,
    ):
        get_assume_role_policy_patch.return_value = self.policy_document
        get_oidc_issuer_id_patch.return_value = self.oidc_provider
        get_account_id_patch.return_value = self.account_id
        self.command += ' --region cn-north-1'

        output = self.run_cmd(self.command, expected_rc=0)
        self.assert_trust_policy_updated(
            output[0], self.expected_policy_document_cn
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import signal
from unittest.mock import patch

import pytest
from botocore import model

import awscli.paramfile
from awscli import shorthand
from awscli.testutils import skip_if_windows, unittest

PARSING_TEST_CASES = (
    # Key val pairs with scalar value.
    ('foo=bar', {'foo': 'bar'}),
    ('foo=bar', {'foo': 'bar'}),
    ('foo=bar,baz=qux', {'foo': 'bar', 'baz': 'qux'}),
    ('a=b,c=d,e=f', {'a': 'b', 'c': 'd', 'e': 'f'}),
    # Empty values are allowed.
    ('foo=', {'foo': ''}),
    ('foo=,bar=', {'foo': '', 'bar': ''}),
    # Unicode is allowed.
    ('foo=\u2713', {'foo': '\u2713'}),
    ('foo=\u2713,\u2713', {'foo': ['\u2713', '\u2713']}),
    # Key val pairs with csv values.
    ('foo=a,b', {'foo': ['a', 'b']}),
    ('foo=a,b,c', {'foo': ['a', 'b', 'c']}),
    ('foo=a,b,bar=c,d', {'foo': ['a', 'b'], 'bar': ['c', 'd']}),
    ('foo=a,b,c,bar=d,e,f', {'foo': ['a', 'b', 'c'], 'bar': ['d', 'e', 'f']}),
    # Spaces in values are allowed.
    ('foo=a,b=with space', {'foo': 'a', 'b': 'with space'}),
    # Trailing spaces are still ignored.
    (
        'foo=a,b=with trailing space  ',
        {'foo': 'a', 'b': 'with trailing space'},
    ),
    ('foo=first space', {'foo': 'first space'}),
    (
        'foo=a space,bar=a space,baz=a space',
        {'foo': 'a space', 'bar': 'a space', 'baz': 'a space'},
    ),
    # Dashes are allowed in key names.
    ('with-dash=bar', {'with-dash': 'bar'}),
    # Underscore are also allowed.
    ('with_underscore=bar', {'with_underscore': 'bar'}),
    # Dots are allowed.
    ('with.dot=bar', {'with.dot': 'bar'}),
    # Pound signs are allowed.
    ('#key=value', {'#key': 'value'}),
    # Forward slashes are allowed in keys.
    ('some/thing=value', {'some/thing': 'value'}),
    # Colon chars are allowed in keys:
    (
        'aws:service:region:124:foo/bar=baz',
        {'aws:service:region:124:foo/bar': 'baz'},
    ),
    # Explicit lists.
    ('foo=[]', {'foo': []}),
    ('foo=[a]', {'foo': ['a']}),
    ('foo=[a,b]', {'foo': ['a', 'b']}),
    ('foo=[a,b,c]', {'foo': ['a', 'b', 'c']}),
    ('foo=[a,b],bar=c,d', {'foo': ['a', 'b'], 'bar': ['c', 'd']}),
    ('foo=[a,b],bar=[c,d]', {'foo': ['a', 'b'], 'bar': ['c', 'd']}),
    ('foo=a,b,bar=[c,d]', {'foo': ['a', 'b'], 'bar': ['c', 'd']}),
    ('foo=[a=b,c=d]', {'foo': ['a=b', 'c=d']}),
    ('foo=[a=b,c=d]', {'foo': ['a=b', 'c=d']}),
    # Lists with whitespace.
    ('foo=[ a , b  , c  ]', {'foo': ['a', 'b', 'c']}),
    ('foo  =  [ a , b  , c  ]', {'foo': ['a', 'b', 'c']}),
    ('foo=[,,]', {'foo': ['', '']}),
    # Single quoted strings.
    ("foo='bar'", {"foo": "bar"}),
    ("foo='bar,baz'", {"foo": "bar,baz"}),
    # Single quoted strings for each value in a CSV list.
    ("foo='bar','baz'", {"foo": ['bar', 'baz']}),
    # Can mix single quoted and non quoted values.
    ("foo=bar,'baz'", {"foo": ['bar', 'baz']}),
    # Quoted strings can include chars not allowed in unquoted strings.
    ("foo=bar,'baz=qux'", {"foo": ['bar', 'baz=qux']}),
    ("foo=bar,'--option=bar space'", {"foo": ['bar', '--option=bar space']}),
    # Can escape the single quote.
    ("foo='bar\\'baz'", {"foo": "bar'baz"}),
    ("foo='bar\\\\baz'", {"foo": "bar\\baz"}),
    # Double quoted strings.
    ('foo="bar"', {'foo': 'bar'}),
    ('foo="bar,baz"', {'foo': 'bar,baz'}),
    ('foo="bar","baz"', {'foo': ['bar', 'baz']}),
    ('foo=bar,"baz=qux"', {'foo': ['bar', 'baz=qux']}),
    ('foo=bar,"--option=bar space"', {'foo': ['bar', '--option=bar space']}),
    ('foo="bar\\"baz"', {'foo': 'bar"baz'}),
    ('foo="bar\\\\baz"', {'foo': 'bar\\baz'}),
    # Can escape comma in CSV list.
    ('foo=a\\,b', {"foo": "a,b"}),
    ('foo=a\\,b', {"foo": "a,b"}),
    ('foo=a\\,', {"foo": "a,"}),
    ('foo=\\,', {"foo": ","}),
    ('foo=a,b\\,c', {"foo": ['a', 'b,c']}),
    ('foo=a,b\\,', {"foo": ['a', 'b,']}),
    ('foo=a,\\,bc', {"foo": ['a', ',bc']}),
    # Ignores whitespace around '=' and ','
    ('foo= bar', {'foo': 'bar'}),
    ('foo =bar', {'foo': 'bar'}),
    ('foo = bar', {'foo': 'bar'}),
    ('foo  =   bar', {'foo': 'bar'}),
    ('foo = bar,baz = qux', {'foo': 'bar', 'baz': 'qux'}),
    ('a = b,  c = d , e = f', {'a': 'b', 'c': 'd', 'e': 'f'}),
    ('foo = ', {'foo': ''}),
    ('a=b,c=  d,  e,  f', {'a': 'b', 'c': ['d', 'e', 'f']}),
    (
        'Name=foo,Values=  a  ,  b  ,  c  ',
        {'Name': 'foo', 'Values': ['a', 'b', 'c']},
    ),
    (
        'Name=foo,Values= a,  b  ,  c',
        {'Name': 'foo', 'Values': ['a', 'b', 'c']},
    ),
    # Can handle newlines between values.
    ('Name=foo,\nValues=a,b,c', {'Name': 'foo', 'Values': ['a', 'b', 'c']}),
    ('A=b,\nC=d,\nE=f\n', {'A': 'b', 'C': 'd', 'E': 'f'}),
    # Hashes
    ('Name={foo=bar,baz=qux}', {'Name': {'foo': 'bar', 'baz': 'qux'}}),
    (
        'Name={foo=[a,b,c],bar=baz}',
        {'Name': {'foo': ['a', 'b', 'c'], 'bar': 'baz'}},
    ),
    ('Name={foo=bar},Bar=baz', {'Name': {'foo': 'bar'}, 'Bar': 'baz'}),
    ('Bar=baz,Name={foo=bar}', {'Bar': 'baz', 'Name': {'foo': 'bar'}}),
    ('a={b={c=d}}', {'a': {'b': {'c': 'd'}}}),
    ('a={b={c=d,e=f},g=h}', {'a': {'b': {'c': 'd', 'e': 'f'}, 'g': 'h'}}),
    # Combining lists and hashes.
    (
        'Name=[{foo=bar}, {baz=qux}]',
        {'Name': [{'foo': 'bar'}, {'baz': 'qux'}]},
    ),
    # Combining hashes and lists.
    (
        'Name=[{foo=[a,b]}, {bar=[c,d]}]',
        {'Name': [{'foo': ['a', 'b']}, {'bar': ['c', 'd']}]},
    ),
    # key-value pairs using @= syntax
    ('foo@=bar', {'foo': 'bar'}),
    ('foo@=bar,baz@=qux', {'foo': 'bar', 'baz': 'qux'}),
    ('foo@=,bar@=', {'foo': '', 'bar': ''}),
    ('foo@=\u2713,\u2713', {'foo': ['\u2713', '\u2713']}),
    ('foo@=a,b,bar=c,d', {'foo': ['a', 'b'], 'bar': ['c', 'd']}),
    ('foo=a,b@=with space', {'foo': 'a', 'b': 'with space'}),
    (
        'foo=a,b@=with trailing space  ',
        {'foo': 'a', 'b': 'with trailing space'},
    ),
    (
        'aws:service:region:124:foo/bar@=baz',
        {'aws:service:region:124:foo/bar': 'baz'},
    ),
    ('foo=[a,b],bar@=[c,d]', {'foo': ['a', 'b'], 'bar': ['c', 'd']}),
    ('foo  @=  [ a , b  , c  ]', {'foo': ['a', 'b', 'c']}),
    ('A=b,\nC@=d,\nE@=f\n', {'A': 'b', 'C': 'd', 'E': 'f'}),
    ('Bar@=baz,Name={foo@=bar}', {'Bar': 'baz', 'Name': {'foo': 'bar'}}),
    (
        'Name=[{foo@=bar}, {baz=qux}]',
        {'Name': [{'foo': 'bar'}, {'baz': 'qux'}]},
    ),
    (
        'Name=[{foo@=[a,b]}, {bar=[c,d]}]',
        {'Name': [{'foo': ['a', 'b']}, {'bar': ['c', 'd']}]},
    ),
)


@pytest.mark.parametrize(
    "expr",
    (
        'foo',
        # Missing closing quotes
        'foo="bar',
        '"foo=bar',
        "foo='bar",
        "foo=[bar",
        "foo={bar",
        "foo={bar}",
        "foo={bar=bar",
        "foo=bar,",
        "foo==bar,\nbar=baz",
        # Duplicate keys should error otherwise they silently
        # set only one of the values.
        'foo=bar,foo=qux',
    ),
)
def test_error_parsing(expr):
    with pytest.raises(shorthand.ShorthandParseError):
        shorthand.ShorthandParser().parse(expr)


@pytest.mark.parametrize(
    "expr",
    (
        # starting with " but unclosed, then repeated \
        'foo="' + '\\' * 100,
        # starting with ' but unclosed, then repeated \
        'foo=\'' + '\\' * 100,
    ),
)
@skip_if_windows("Windows does not support signal.SIGALRM.")
def test_error_with_backtracking(expr):
    signal.signal(signal.SIGALRM, handle_timeout)
    # Ensure we don't spend more than 5 seconds backtracking
    signal.alarm(5)
    with pytest.raises(shorthand.ShorthandParseError):
        shorthand.ShorthandParser().parse(expr)
    signal.alarm(0)


def handle_timeout(signum, frame):
    raise TimeoutError('Shorthand parsing timed out')


@pytest.mark.parametrize('data, expected', PARSING_TEST_CASES)
def test_parse(data, expected):
    actual = shorthand.ShorthandParser().parse(data)
    assert actual == expected


class TestShorthandParserParamFile:
    @patch('awscli.paramfile.compat_open')
    @pytest.mark.parametrize(
        'file_contents, data, expected',
        (
            (
                'file-contents123',
                'Foo@=file://foo,Bar={Baz@=file://foo}',
                {
                    'Foo': 'file-contents123',
                    'Bar': {'Baz': 'file-contents123'},
                },
            ),
            (
                b'file-contents123',
                'Foo@=fileb://foo,Bar={Baz@=fileb://foo}',
                {
                    'Foo': b'file-contents123',
                    'Bar': {'Baz': b'file-contents123'},
                },
            ),
            (
                'file-contents123',
                'Bar@={Baz=file://foo}',
                {'Bar': {'Baz': 'file://foo'}},
            ),
            (
                'file-contents123',
                'Foo@=foo,Bar={Baz@=foo}',
                {'Foo': 'foo', 'Bar': {'Baz': 'foo'}},
            ),
        ),
    )
    def test_paramfile(self, mock_compat_open, file_contents, data, expected):
        mock_compat_open.return_value.__enter__.return_value.read.return_value = file_contents
        result = shorthand.ShorthandParser().parse(data)
        assert result == expected

    @patch('awscli.paramfile.compat_open')
    def test_paramfile_list(self, mock_compat_open):
        f1_contents = 'file-contents123'
        f2_contents = 'contents2'
        mock_compat_open.return_value.__enter__.return_value.read.side_effect = [
            f1_contents,
            f2_contents,
        ]
        result = shorthand.ShorthandParser().parse(
            'Foo@=[a, file://foo1, file://foo2]'
        )
        assert result == {'Foo': ['a', f1_contents, f2_contents]}

    def test_paramfile_does_not_exist_error(self, capsys):
        with pytest.raises(awscli.paramfile.ResourceLoadingError):
            shorthand.ShorthandParser().parse('Foo@=file://fakefile.txt')
            captured = capsys.readouterr()
            assert "No such file or directory: 'fakefile.txt" in captured.err


class TestModelVisitor(unittest.TestCase):
    def test_promote_to_list_of_ints(self):
        m = (
            model.DenormalizedStructureBuilder()
            .with_members(
                {
                    'A': {'type': 'list', 'member': {'type': 'string'}},
                }
            )
            .build_model()
        )
        b = shorthand.BackCompatVisitor()

        params = {'A': 'foo'}
        b.visit(params, m)
        self.assertEqual(params, {'A': ['foo']})

    def test_dont_promote_list_if_none_value(self):
        m = (
            model.DenormalizedStructureBuilder()
            .with_members(
                {
                    'A': {
                        'type': 'list',
                        'member': {
                            'type': 'structure',
                            'members': {'Single': {'type': 'string'}},
                        },
                    },
                }
            )
            .build_model()
        )
        b = shorthand.BackCompatVisitor()
        params = {}
        b.visit(params, m)
        self.assertEqual(params, {})

    def test_can_convert_scalar_types_from_string(self):
        m = (
            model.DenormalizedStructureBuilder()
            .with_members(
                {
                    'A': {'type': 'integer'},
                    'B': {'type': 'string'},
                    'C': {'type': 'float'},
                    'D': {'type': 'boolean'},
                    'E': {'type': 'boolean'},
                }
            )
            .build_model()
        )
        b = shorthand.BackCompatVisitor()

        params = {
            'A': '24',
            'B': '24',
            'C': '24.12345',
            'D': 'true',
            'E': 'false',
        }
        b.visit(params, m)
        self.assertEqual(
            params,
            {
                'A': 24,
                'B': '24',
                'C': float('24.12345'),
                'D': True,
                'E': False,
            },
        )

    def test_empty_values_not_added(self):
        m = (
            model.DenormalizedStructureBuilder()
            .with_members(
                {
                    'A': {'type': 'boolean'},
                }
            )
            .build_model()
        )
        b = shorthand.BackCompatVisitor()

        params = {}
        b.visit(params, m)
        self.assertEqual(params, {})

    def test_can_convert_list_of_integers(self):
        m = (
            model.DenormalizedStructureBuilder()
            .with_members(
                {
                    'A': {
                        'type': 'list',
                        'member': {
                            'type': 'integer',
                        },
                    },
                }
            )
            .build_model()
        )
        b = shorthand.BackCompatVisitor()
        params = {'A': ['1', '2']}
        b.visit(params, m)
        # We should have converted each list element to an integer
        # because the type of the list member is integer.
        self.assertEqual(params, {'A': [1, 2]})



# Copyright (c) 2013 Amazon.com, Inc. or its affiliates.  All Rights Reserved
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
import sys

from awscli import text
from awscli.compat import StringIO
from awscli.testutils import mock, unittest


class TestSection(unittest.TestCase):
    def format_text(self, data, stream=None):
        if stream is None:
            stream = StringIO()
        text.format_text(data, stream=stream)
        return stream.getvalue()

    def assert_text_renders_to(self, data, expected_rendering):
        rendered = self.format_text(data)
        self.assertEqual(rendered, expected_rendering)

    def test_dict_format(self):
        self.assert_text_renders_to(dict(a=1, b=2, c=3), "1\t2\t3\n")

    def test_list_format(self):
        self.assert_text_renders_to([1, 2, 3], "1\t2\t3\n")

    def test_list_of_dicts(self):
        self.assert_text_renders_to(
            {'foo': [dict(a=1, b=2, c=3), dict(a=4, b=5, c=6)]},
            'FOO\t1\t2\t3\n' 'FOO\t4\t5\t6\n',
        )

    def test_multiple_list_of_dicts(self):
        self.assert_text_renders_to(
            {
                'foo': [dict(a=1, b=2, c=3), dict(a=4, b=5, c=6)],
                'zoo': [dict(a=7, b=8, c=9), dict(a=0, b=1, c=2)],
            },
            'FOO\t1\t2\t3\n'
            'FOO\t4\t5\t6\n'
            'ZOO\t7\t8\t9\n'
            'ZOO\t0\t1\t2\n',
        )

    def test_single_scalar_number(self):
        self.assert_text_renders_to(10, '10\n')

    def test_list_of_single_number(self):
        self.assert_text_renders_to([10], '10\n')

    def test_list_of_multiple_numbers(self):
        self.assert_text_renders_to([10, 10, 10], '10\t10\t10\n')

    def test_different_keys_in_sublists(self):
        self.assert_text_renders_to(
            #                                missing "b"        adds "d"
            {'foo': [dict(a=1, b=2, c=3), dict(a=4, c=5), dict(a=6, d=7)]},
            'FOO\t1\t2\t3\t\n' 'FOO\t4\t\t5\t\n' 'FOO\t6\t\t\t7\n',
        )

    def test_different_keys_in_nested_sublists(self):
        self.assert_text_renders_to(
            {
                'bar': [
                    {'foo': [dict(a=1, b=2, c=3), dict(a=4, c=5)]},
                    {'foo': [dict(b=6, d=7), dict(b=8, c=9)]},
                ]
            },
            'FOO\t1\t2\t3\n' 'FOO\t4\t\t5\n' 'FOO\t6\t\t7\n' 'FOO\t8\t9\t\n',
        )

    def test_different_keys_in_deeply_nested_sublists(self):
        self.assert_text_renders_to(
            {
                'bar': [
                    {'foo': [[[dict(a=1, b=2, c=3), dict(a=4, c=5)]]]},
                    {'foo': [[[dict(b=6, d=7), dict(b=8, c=9)]]]},
                ]
            },
            'FOO\t1\t2\t3\n' 'FOO\t4\t\t5\n' 'FOO\t6\t\t7\n' 'FOO\t8\t9\t\n',
        )

    def test_scalars_and_complex_types(self):
        self.assert_text_renders_to(
            {
                'foo': [
                    dict(a=1, b=dict(y='y', z='z'), c=3),
                    dict(a=4, b=dict(y='y', z='z'), c=6),
                ]
            },
            'FOO\t1\t3\n' 'B\ty\tz\n' 'FOO\t4\t6\n' 'B\ty\tz\n',
        )

    def test_nested_list_of_lists(self):
        self.assert_text_renders_to(
            [['1', '2', '3'], ['4', '5', '6']], '1\t2\t3\n' '4\t5\t6\n'
        )

    def test_deeply_nested_lists(self):
        self.assert_text_renders_to(
            [
                [['1', '2', '3'], ['4', '5', '6']],
                [['7', '8', '9'], ['0', '1', '2']],
            ],
            '1\t2\t3\n' '4\t5\t6\n' '7\t8\t9\n' '0\t1\t2\n',
        )

    def test_unicode_text(self):
        self.assert_text_renders_to([['1', '2', '\u2713']], '1\t2\t\u2713\n')

    def test_single_scalar_value(self):
        self.assert_text_renders_to('foobarbaz', 'foobarbaz\n')

    def test_empty_list(self):
        self.assert_text_renders_to([], '')

    def test_empty_inner_list(self):
        self.assert_text_renders_to([[]], '')

    def test_deeploy_nested_empty_list(self):
        self.assert_text_renders_to([[[[]]]], '')

    def test_deeploy_nested_single_scalar(self):
        self.assert_text_renders_to([[[['a']]]], 'a\n')

    def test_empty_list_mock_calls(self):
        # We also need this test as well as test_empty_list
        # because we want to ensure that write() is never called with
        # a list object.
        fake_stream = mock.Mock()
        self.format_text(data=[], stream=fake_stream)
        # We should not call .write() at all for an empty list.
        self.assertFalse(fake_stream.write.called)

    def test_list_of_strings_in_dict(self):
        self.assert_text_renders_to(
            {'KeyName': ['a', 'b', 'c']},
            'KEYNAME\ta\n' 'KEYNAME\tb\n' 'KEYNAME\tc\n',
        )

    def test_inconsistent_sublists(self):
        self.assert_text_renders_to(
            [[['1', '2'], ['3', '4', '5', '6']], [['7', '8', '9'], ['0']]],
            '1\t2\n' '3\t4\t5\t6\n' '7\t8\t9\n' '0\n',
        )

    def test_lists_mixed_with_scalars(self):
        self.assert_text_renders_to(
            [['a', 'b', ['c', 'd']], ['e', 'f', ['g', 'h']]],
            'a\tb\n' 'c\td\n' 'e\tf\n' 'g\th\n',
        )

    def test_deeply_nested_with_scalars(self):
        self.assert_text_renders_to(
            [
                ['a', 'b', ['c', 'd', ['e', 'f', ['g', 'h']]]],
                ['i', 'j', ['k', 'l', ['m', 'n', ['o', 'p']]]],
            ],
            'a\tb\n'
            'c\td\n'
            'e\tf\n'
            'g\th\n'
            'i\tj\n'
            'k\tl\n'
            'm\tn\n'
            'o\tp\n',
        )

    def test_deeply_nested_with_identifier(self):
        self.assert_text_renders_to(
            {'foo': [['a', 'b', ['c', 'd']], ['e', 'f', ['g', 'h']]]},
            'FOO\ta\n'
            'FOO\tb\n'
            'FOO\tc\n'
            'FOO\td\n'
            'FOO\te\n'
            'FOO\tf\n'
            'FOO\tg\n'
            'FOO\th\n',
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.session import Session

from awscli.customizations.wizard.devcommands import WizardDev, WizardDevRunner
from awscli.testutils import mock, unittest


class TestWizardDevCommand(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock(spec=Session)
        self.session.emit_first_non_none_response.return_value = None
        self.dev_runner = mock.Mock(spec=WizardDevRunner)

    def test_will_delegate_to_runner(self):
        cmd = WizardDev(self.session, dev_runner=self.dev_runner)
        cmd(['--run-wizard', 'mywizard.yml'], parsed_globals=None)
        self.dev_runner.run_wizard.assert_called_with('mywizard.yml')



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from awscli.testutils import BaseAWSCommandParamsTest


class TestGenerateCliSkeletonInput(BaseAWSCommandParamsTest):
    def test_generate_cli_skeleton_s3api(self):
        cmdline = 's3api delete-object --generate-cli-skeleton'
        stdout, _, rc = self.run_cmd(cmdline)
        self.assertEqual(rc, 0)
        loaded_skeleton = json.loads(stdout)
        self.assertEqual(loaded_skeleton['Bucket'], '')
        self.assertEqual(loaded_skeleton['BypassGovernanceRetention'], True)
        self.assertEqual(loaded_skeleton['Key'], '')
        self.assertEqual(loaded_skeleton['MFA'], '')
        self.assertEqual(loaded_skeleton['RequestPayer'], 'requester')

    def test_generate_cli_skeleton_sqs(self):
        cmdline = 'sqs change-message-visibility --generate-cli-skeleton'
        stdout, _, rc = self.run_cmd(cmdline)
        self.assertEqual(rc, 0)
        loaded_skeleton = json.loads(stdout)
        self.assertEqual(loaded_skeleton['QueueUrl'], '')
        self.assertEqual(loaded_skeleton['ReceiptHandle'], '')
        self.assertEqual(loaded_skeleton['VisibilityTimeout'], 0)

    def test_generate_cli_skeleton_iam(self):
        cmdline = 'iam create-group --generate-cli-skeleton'
        stdout, _, rc = self.run_cmd(cmdline)
        self.assertEqual(rc, 0)
        loaded_skeleton = json.loads(stdout)
        self.assertEqual(loaded_skeleton['Path'], '')
        self.assertEqual(loaded_skeleton['GroupName'], '')


class TestGenerateCliSkeletonYamlInput(BaseAWSCommandParamsTest):
    def test_generate_cli_skeleton_s3api(self):
        cmdline = 's3api delete-object --generate-cli-skeleton yaml-input'
        stdout, _, rc = self.run_cmd(cmdline)
        self.assertEqual(rc, 0)
        loaded_skeleton = self.yaml.load(stdout)
        self.assertEqual(loaded_skeleton['Bucket'], '')
        self.assertEqual(loaded_skeleton['BypassGovernanceRetention'], True)
        self.assertEqual(loaded_skeleton['Key'], '')
        self.assertEqual(loaded_skeleton['MFA'], '')
        self.assertEqual(loaded_skeleton['RequestPayer'], 'requester')

    def test_generate_cli_skeleton_sqs(self):
        stdout, _, rc = self.run_cmd(
            'sqs change-message-visibility --generate-cli-skeleton yaml-input'
        )
        self.assertEqual(rc, 0)
        loaded_skeleton = self.yaml.load(stdout)
        self.assertEqual(loaded_skeleton['QueueUrl'], '')
        self.assertEqual(loaded_skeleton['ReceiptHandle'], '')
        self.assertEqual(loaded_skeleton['VisibilityTimeout'], 0)

    def test_generate_cli_skeleton_iam(self):
        cmdline = 'iam create-group --generate-cli-skeleton yaml-input'
        stdout, _, rc = self.run_cmd(cmdline)
        self.assertEqual(rc, 0)
        loaded_skeleton = self.yaml.load(stdout)
        self.assertEqual(loaded_skeleton['Path'], '')
        self.assertEqual(loaded_skeleton['GroupName'], '')

    def test_contains_comments(self):
        # Choose a command that we know will have comments as there are
        # required parameters and required parameters always are labeled
        # with required.
        cmdline = 's3api delete-object --generate-cli-skeleton yaml-input'
        stdout, _, rc = self.run_cmd(cmdline)
        self.assertIn("Bucket: ''  # [REQUIRED]", stdout)


class TestGenerateCliSkeletonOutput(BaseAWSCommandParamsTest):
    def test_generate_cli_skeleton_output(self):
        cmdline = 'ec2 describe-regions --generate-cli-skeleton output'
        stdout, _, _ = self.run_cmd(cmdline)
        # The format of the response should be a json blob with the
        # following structure:
        # {
        #     "Regions": [
        #         {
        #             "RegionName": "RegionName",
        #             "Endpoint": "Endpoint"
        #         }
        #     ]
        # }
        #
        # We assert only components of the response in case members
        # are added in the future that would break an exactly equals
        # assertion
        skeleton_output = json.loads(stdout)
        self.assertIn('Regions', skeleton_output)
        self.assertEqual(
            skeleton_output['Regions'][0]['RegionName'], 'RegionName'
        )
        self.assertEqual(skeleton_output['Regions'][0]['Endpoint'], 'Endpoint')

    def test_can_pass_in_input_parameters(self):
        cmdline = 'ec2 describe-regions --generate-cli-skeleton output '
        cmdline += ' --region-names us-east-1'
        stdout, _, _ = self.assert_params_for_cmd(
            cmdline, {'RegionNames': ['us-east-1']}
        )

        # Make sure the output has the proper mocked response as well.
        skeleton_output = json.loads(stdout)
        self.assertIn('Regions', skeleton_output)
        self.assertEqual(
            skeleton_output['Regions'][0]['RegionName'], 'RegionName'
        )
        self.assertEqual(skeleton_output['Regions'][0]['Endpoint'], 'Endpoint')

    def test_when_no_output_shape(self):
        cmdline = 'ec2 attach-internet-gateway '
        cmdline += '--internet-gateway-id igw-c0a643a9 --vpc-id vpc-a01106 '
        cmdline += '--generate-cli-skeleton output'
        stdout, _, _ = self.assert_params_for_cmd(
            cmdline,
            {'InternetGatewayId': 'igw-c0a643a9', 'VpcId': 'vpc-a01106'},
        )
        # There should be no output as the command has no output shape
        self.assertEqual('', stdout)

    def test_can_handle_timestamps(self):
        cmdline = 's3api list-buckets --generate-cli-skeleton output'
        stdout, _, _ = self.run_cmd(cmdline)
        skeleton_output = json.loads(stdout)
        # The CreationDate has the type of timestamp
        self.assertEqual(
            skeleton_output['Buckets'][0]['CreationDate'],
            '1970-01-01T00:00:00',
        )

    def test_can_handle_lists_with_strings_that_have_a_min_length(self):
        cmdline = 'dynamodb list-tables --generate-cli-skeleton output'
        stdout, _, _ = self.run_cmd(cmdline)
        skeleton_output = json.loads(stdout)
        self.assertEqual(skeleton_output['TableNames'], ['TableName'])

    def test_respects_formatting(self):
        cmdline = 'ec2 describe-regions --generate-cli-skeleton output '
        cmdline += ' --query Regions[].RegionName --output text'
        stdout, _, _ = self.run_cmd(cmdline)
        self.assertEqual(stdout, 'RegionName\n')

    def test_validates_at_command_line_level(self):
        cmdline = 'ec2 start-instances --generate-cli-skeleton output'
        stdout, stderr, _ = self.run_cmd(cmdline, expected_rc=252)
        self.assertIn('required', stderr)
        self.assertIn('--instance-ids', stderr)
        self.assertEqual('', stdout)

    def test_validates_at_client_level(self):
        cmdline = 'ec2 describe-instances --generate-cli-skeleton output '
        # Note: The for --filters instead of Value the key should be Values
        # which should throw a validation error.
        cmdline += '--filters Name=instance-id,Value=foo'
        stdout, stderr, _ = self.run_cmd(cmdline, expected_rc=252)
        self.assertIn('Unknown parameter in Filters[0]', stderr)
        self.assertEqual('', stdout)



import re

import pytest

from awscli.customizations.ecs.expressgateway.managedresource import (
    TERMINAL_RESOURCE_STATUSES,
    ManagedResource,
)


class TestManagedResource:
    def test_is_terminal_active(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        assert resource.is_terminal()

    def test_is_terminal_failed(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "FAILED", 1761230543.151
        )
        assert resource.is_terminal()

    def test_is_terminal_provisioning(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "PROVISIONING", 1761230543.151
        )
        assert not resource.is_terminal()

    def test_get_status_string_active(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        status_string = resource.get_status_string("⠋")
        assert "LoadBalancer" in status_string
        assert "lb-123" in status_string
        assert "ACTIVE" in status_string

    def test_get_status_string_failed_with_reason(self):
        resource = ManagedResource(
            "LoadBalancer",
            "lb-123",
            "FAILED",
            1761230543.151,
            "Connection timeout",
        )
        status_string = resource.get_status_string("⠋")
        assert "FAILED" in status_string
        assert "Connection timeout" in status_string

    def test_get_status_string_active_with_reason(self):
        resource = ManagedResource(
            "LoadBalancer",
            "lb-123",
            "ACTIVE",
            1761230543.151,
            "Load balancer ready",
        )
        status_string = resource.get_status_string("⠋")
        assert "ACTIVE" in status_string
        assert "Load balancer ready" in status_string

    def test_combine_newer_resource(self):
        older = ManagedResource(
            "LoadBalancer", "lb-123", "PROVISIONING", 1761230543.151
        )
        newer = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230600.151
        )
        result = older.combine(newer)
        assert result == newer

    def test_combine_older_resource(self):
        older = ManagedResource(
            "LoadBalancer", "lb-123", "PROVISIONING", 1761230543.151
        )
        newer = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230600.151
        )
        result = newer.combine(older)
        assert result == newer

    def test_combine_with_none(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        result = resource.combine(None)
        assert result == resource

    def test_is_terminal_deleted_status(self):
        resource = ManagedResource("LoadBalancer", "lb-123", "DELETED")
        assert resource.is_terminal()

    def test_is_terminal_no_status(self):
        resource = ManagedResource("LoadBalancer", "lb-123", None)
        assert not resource.is_terminal()

    def test_init_with_string_timestamp(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", "2025-11-05T18:00:00Z"
        )
        assert isinstance(resource.updated_at, float)
        assert resource.updated_at > 0

    def test_init_with_none_timestamp(self):
        resource = ManagedResource("LoadBalancer", "lb-123", "ACTIVE", None)
        assert resource.updated_at is None

    def test_combine_with_no_timestamp(self):
        resource1 = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        resource2 = ManagedResource(
            "LoadBalancer", "lb-123", "PROVISIONING", None
        )
        result = resource1.combine(resource2)
        assert result == resource1

    def test_combine_equal_timestamps(self):
        timestamp = 1761230543.151
        resource1 = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", timestamp
        )
        resource2 = ManagedResource(
            "LoadBalancer", "lb-123", "PROVISIONING", timestamp
        )
        result = resource1.combine(resource2)
        assert result == resource1

    def test_get_status_string_with_depth(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        status_string = resource.get_status_string("⠋", depth=2)
        # Should have proper indentation
        lines = status_string.split('\n')
        assert lines[0].startswith("  ")  # 2 spaces for depth=2

    def test_get_status_string_with_additional_info(self):
        resource = ManagedResource(
            "LoadBalancer",
            "lb-123",
            "ACTIVE",
            1761230543.151,
            additional_info="Load balancer is healthy",
        )
        status_string = resource.get_status_string("⠋")
        assert "Load balancer is healthy" in status_string

    def test_get_status_string_no_identifier(self):
        resource = ManagedResource(
            "LoadBalancer", None, "ACTIVE", 1761230543.151
        )
        status_string = resource.get_status_string("⠋")
        assert "LoadBalancer" in status_string
        assert "ACTIVE" in status_string

    def test_get_status_string_no_color(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        status_string = resource.get_status_string("⠋", use_color=False)
        assert "LoadBalancer" in status_string
        assert "lb-123" in status_string
        assert "ACTIVE" in status_string
        # Should not contain ANSI color codes
        assert "\x1b[" not in status_string

    def test_get_status_string_with_color(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        status_string = resource.get_status_string("⠋", use_color=True)
        assert "LoadBalancer" in status_string
        assert "lb-123" in status_string
        assert "ACTIVE" in status_string
        # Should contain ANSI color codes
        assert "\x1b[" in status_string

    def test_get_stream_string_basic(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        stream_string = resource.get_stream_string("2025-12-15 10:00:00")
        assert "[2025-12-15 10:00:00]" in stream_string
        assert "LoadBalancer" in stream_string
        assert "lb-123" in stream_string
        assert "ACTIVE" in stream_string

    def test_get_stream_string_with_reason(self):
        resource = ManagedResource(
            "LoadBalancer",
            "lb-123",
            "PROVISIONING",
            1761230543.151,
            "Waiting for DNS propagation",
        )
        stream_string = resource.get_stream_string("2025-12-15 10:00:00")
        assert "Reason: Waiting for DNS propagation" in stream_string

    def test_get_stream_string_with_additional_info(self):
        resource = ManagedResource(
            "LoadBalancer",
            "lb-123",
            "ACTIVE",
            1761230543.151,
            additional_info="DNS: example.elb.amazonaws.com",
        )
        stream_string = resource.get_stream_string("2025-12-15 10:00:00")
        assert "Info: DNS: example.elb.amazonaws.com" in stream_string

    def test_get_stream_string_with_updated_at(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        stream_string = resource.get_stream_string("2025-12-15 10:00:00")
        assert "Last Updated At:" in stream_string
        # Check timestamp format YYYY-MM-DD HH:MM:SS
        assert re.search(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}", stream_string)

    def test_get_stream_string_all_fields(self):
        resource = ManagedResource(
            "TargetGroup",
            "tg-456",
            "PROVISIONING",
            1761230543.151,
            "Registering targets",
            "Health check interval: 30s",
        )
        stream_string = resource.get_stream_string("2025-12-15 10:00:00")
        assert "[2025-12-15 10:00:00]" in stream_string
        assert "TargetGroup" in stream_string
        assert "tg-456" in stream_string
        assert "PROVISIONING" in stream_string
        assert "Reason: Registering targets" in stream_string
        assert "Last Updated At:" in stream_string
        assert "Info: Health check interval: 30s" in stream_string

    def test_get_stream_string_no_identifier(self):
        resource = ManagedResource(
            "LoadBalancer", None, "ACTIVE", 1761230543.151
        )
        stream_string = resource.get_stream_string("2025-12-15 10:00:00")
        assert "[2025-12-15 10:00:00]" in stream_string
        assert "LoadBalancer" in stream_string
        assert "ACTIVE" in stream_string
        assert "None" not in stream_string

    def test_get_stream_string_no_status(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", None, 1761230543.151
        )
        stream_string = resource.get_stream_string("2025-12-15 10:00:00")
        assert "LoadBalancer" in stream_string
        assert "lb-123" in stream_string
        # Should not have status brackets when status is None
        assert "[None]" not in stream_string

    def test_get_stream_string_no_color(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        stream_string = resource.get_stream_string(
            "2025-12-15 10:00:00", use_color=False
        )
        assert "LoadBalancer" in stream_string
        assert "lb-123" in stream_string
        assert "ACTIVE" in stream_string
        # Should not contain ANSI color codes
        assert "\x1b[" not in stream_string

    def test_get_stream_string_with_color(self):
        resource = ManagedResource(
            "LoadBalancer", "lb-123", "ACTIVE", 1761230543.151
        )
        stream_string = resource.get_stream_string(
            "2025-12-15 10:00:00", use_color=True
        )
        assert "LoadBalancer" in stream_string
        assert "lb-123" in stream_string
        assert "ACTIVE" in stream_string
        # Should contain ANSI color codes
        assert "\x1b[" in stream_string

    def test_get_stream_string_failed_status(self):
        resource = ManagedResource(
            "LoadBalancer",
            "lb-123",
            "FAILED",
            1761230543.151,
            "Connection timeout",
        )
        stream_string = resource.get_stream_string("2025-12-15 10:00:00")
        assert "FAILED" in stream_string
        assert "Reason: Connection timeout" in stream_string
        # Failed status should use color coding
        assert "\x1b[" in stream_string


class TestConstants:
    def test_terminal_resource_statuses(self):
        expected_statuses = ["ACTIVE", "DELETED", "FAILED"]
        assert TERMINAL_RESOURCE_STATUSES == expected_statuses



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


import json

from awscli.testutils import mock
from tests.unit.customizations.emr import (
    EMRBaseAWSCommandParamsTest as BaseAWSCommandParamsTest,
)

INSTALL_HIVE_STEP = {
    'HadoopJarStep': {
        'Args': [
            's3://us-east-1.elasticmapreduce/libs/hive/hive-script',
            '--install-hive',
            '--base-path',
            's3://us-east-1.elasticmapreduce/libs/hive',
            '--hive-versions',
            'latest',
        ],
        'Jar': (
            's3://us-east-1.elasticmapreduce/libs/'
            'script-runner/script-runner.jar'
        ),
    },
    'Name': 'Install Hive',
    'ActionOnFailure': 'TERMINATE_CLUSTER',
}

INSTALL_HIVE_SITE_STEP = {
    'HadoopJarStep': {
        'Args': [
            's3://us-east-1.elasticmapreduce/libs/hive/hive-script',
            '--base-path',
            's3://us-east-1.elasticmapreduce/libs/hive',
            '--install-hive-site',
            '--hive-site=s3://test/hive-conf/hive-site.xml',
            '--hive-versions',
            'latest',
        ],
        'Jar': (
            's3://us-east-1.elasticmapreduce/libs/'
            'script-runner/script-runner.jar'
        ),
    },
    'Name': 'Install Hive Site Configuration',
    'ActionOnFailure': 'CANCEL_AND_WAIT',
}

INSTALL_PIG_STEP = {
    'HadoopJarStep': {
        'Args': [
            's3://us-east-1.elasticmapreduce/libs/pig/pig-script',
            '--install-pig',
            '--base-path',
            's3://us-east-1.elasticmapreduce/libs/pig',
            '--pig-versions',
            'latest',
        ],
        'Jar': (
            's3://us-east-1.elasticmapreduce/libs/'
            'script-runner/script-runner.jar'
        ),
    },
    'Name': 'Install Pig',
    'ActionOnFailure': 'TERMINATE_CLUSTER',
}


class TestInstallApplications(BaseAWSCommandParamsTest):
    prefix = (
        'emr install-applications --cluster-id ' 'j-ABC123456 --applications '
    )

    def test_install_hive_site(self):
        cmdline = (
            self.prefix + 'Name=Hive,'
            'Args=[--hive-site=s3://test/hive-conf/hive-site.xml]'
        )
        result = {
            'JobFlowId': 'j-ABC123456',
            'Steps': [INSTALL_HIVE_STEP, INSTALL_HIVE_SITE_STEP],
        }
        self.assert_params_for_cmd(cmdline, result)
        cmdline = (
            self.prefix + 'Name=Hive,'
            'Args=[--hive-site=s3://test/hive-conf/hive-site.xml,k1]'
        )
        self.assert_params_for_cmd(cmdline, result)

    def test_install_hive_and_pig(self):
        cmdline = self.prefix + 'Name=Hive Name=Pig'
        result = {
            'JobFlowId': 'j-ABC123456',
            'Steps': [INSTALL_HIVE_STEP, INSTALL_PIG_STEP],
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_install_pig_with_profile_region(self):
        self.driver.session.set_config_variable('region', 'cn-north-1')
        cmdline = self.prefix + 'Name=Pig'
        PIG_STEP = json.dumps(INSTALL_PIG_STEP).replace(
            'us-east-1', 'cn-north-1'
        )
        result = {'JobFlowId': 'j-ABC123456', 'Steps': [json.loads(PIG_STEP)]}
        self.assert_params_for_cmd(cmdline, result)

    def test_install_impala_error(self):
        cmdline = self.prefix + ' Name=Impala'

        expected_error_msg = (
            "\naws: [ERROR]: An error occurred (ParamValidation): Impala cannot be installed on"
            + " a running cluster. 'Name' should be one of the following:"
            + " HIVE, PIG\n"
        )
        result = self.run_cmd(cmdline, 252)
        self.assertEqual(result[1], expected_error_msg)

    def test_install_unknown_app_error(self):
        cmdline = self.prefix + 'Name=unknown'

        expected_error_msg = (
            "\naws: [ERROR]: An error occurred (ParamValidation): Unknown application: unknown."
            + " 'Name' should be one of the following: HIVE, PIG, HBASE,"
            + " GANGLIA, IMPALA, SPARK, MAPR, MAPR_M3, MAPR_M5, MAPR_M7\n"
        )
        result = self.run_cmd(cmdline, 252)
        self.assertEqual(result[1], expected_error_msg)

    @mock.patch('awscli.customizations.emr.' 'emrutils.get_release_label')
    def test_unsupported_command_on_release_based_cluster_error(
        self, grl_patch
    ):
        grl_patch.return_value = 'emr-4.0'
        cmdline = (
            self.prefix + 'Name=Hive,'
            'Args=[--hive-site=s3://test/hive-conf/hive-site.xml]'
        )

        expected_error_msg = (
            "\naws: [ERROR]: An error occurred (ParamValidation): install-applications"
            " is not supported with 'emr-4.0' release.\n"
        )
        result = self.run_cmd(cmdline, 252)
        self.assertEqual(result[1], expected_error_msg)


if __name__ == "__main__":
    unittest.main()



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.wizard.ui.style import get_default_style
from awscli.testutils import mock, unittest


class TestStyles(unittest.TestCase):
    @mock.patch('awscli.customizations.wizard.ui.style.is_windows')
    def test_get_styles_for_windows(self, is_windows):
        is_windows.return_value = True
        style = get_default_style()
        self.assertIn(('shadow', 'bg:#eeeeee'), style.style_rules)
        self.assertNotIn(('shadow', 'bg:#222222'), style.style_rules)

    @mock.patch('awscli.customizations.wizard.ui.style.is_windows')
    def test_get_styles_for_non_windows(self, is_windows):
        is_windows.return_value = False
        style = get_default_style()
        self.assertIn(('shadow', 'bg:#222222'), style.style_rules)
        self.assertNotIn(('shadow', 'bg:#eeeeee'), style.style_rules)



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest, FileCreator


class TestTranslateDocument(BaseAWSCommandParamsTest):
    prefix = "translate translate-document"

    def setUp(self):
        super().setUp()
        self.files = FileCreator()
        self.temp_file = self.files.create_file("foo", "mycontents")
        with open(self.temp_file, "rb") as f:
            self.temp_file_contents = f.read()

    def tearDown(self):
        super().tearDown()
        self.files.remove_all()

    def test_translate_document_with_file(self):
        cmdline = self.prefix
        cmdline += " --source-language-code FOO"
        cmdline += " --target-language-code BAR"
        cmdline += " --document ContentType=datatype"
        cmdline += " --document-content fileb://%s" % self.temp_file
        result = {
            "Document": {
                "Content": self.temp_file_contents,
                "ContentType": "datatype",
            },
            "SourceLanguageCode": "FOO",
            "TargetLanguageCode": "BAR",
        }
        self.assert_params_for_cmd(cmdline, result)

    def test_translate_document_with_orignal_argument(self):
        cmdline = self.prefix
        cmdline += " --source-language-code FOO"
        cmdline += " --target-language-code BAR"
        cmdline += " --document Content=data,ContentType=datatype"

        stdout, stderr, rc = self.assert_params_for_cmd(
            cmdline, expected_rc=252
        )
        self.assertIn(
            "the following arguments are required: --document-content", stderr
        )

    def test_translate_document_with_file_and_orignal_argument(self):
        cmdline = self.prefix
        cmdline += " --source-language-code FOO"
        cmdline += " --target-language-code BAR"
        cmdline += " --document Content=data,ContentType=datatype"
        cmdline += " --document-content fileb://%s" % self.temp_file

        stdout, stderr, rc = self.assert_params_for_cmd(
            cmdline, expected_rc=252
        )
        self.assertIn(
            "Content cannot be provided as a part of the '--document' "
            "argument. Please use the '--document-content' option instead to "
            "specify a file.",
            stderr,
        )



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.compat import OrderedDict, json

from awscli.customizations.datapipeline import translator
from awscli.customizations.datapipeline.translator import (
    PipelineDefinitionError,
)
from awscli.testutils import unittest

# Throughout these tests, 'df' refers to the condensed JSON definition format
# that the user provides and API refers to the format expected by the API.


class TestTranslatePipelineDefinitions(unittest.TestCase):
    maxDiff = None

    def load_def(self, json_string):
        return json.loads(json_string, object_pairs_hook=OrderedDict)

    def test_convert_schedule_df_to_api(self):
        definition = self.load_def("""{"objects": [
            {
              "id" : "S3ToS3Copy",
              "type" : "CopyActivity",
              "schedule" : { "ref" : "CopyPeriod" },
              "input" : { "ref" : "InputData" },
              "output" : { "ref" : "OutputData" }
            }
            ]}""")
        actual = translator.definition_to_api_objects(definition)
        api = [
            {
                "name": "S3ToS3Copy",
                "id": "S3ToS3Copy",
                "fields": [
                    {"key": "input", "refValue": "InputData"},
                    {"key": "output", "refValue": "OutputData"},
                    {"key": "schedule", "refValue": "CopyPeriod"},
                    {"key": "type", "stringValue": "CopyActivity"},
                ],
            }
        ]
        self.assertEqual(actual, api)

    def test_convert_df_to_api_schedule(self):
        definition = self.load_def("""{
              "objects": [
                {
                  "id": "MySchedule",
                  "type": "Schedule",
                  "startDateTime": "2013-08-18T00:00:00",
                  "endDateTime": "2013-08-19T00:00:00",
                  "period": "1 day"
                }
            ]}""")
        actual = translator.definition_to_api_objects(definition)
        api = [
            {
                "name": "MySchedule",
                "id": "MySchedule",
                "fields": [
                    {
                        "key": "endDateTime",
                        "stringValue": "2013-08-19T00:00:00",
                    },
                    {"key": "period", "stringValue": "1 day"},
                    {
                        "key": "startDateTime",
                        "stringValue": "2013-08-18T00:00:00",
                    },
                    {"key": "type", "stringValue": "Schedule"},
                ],
            }
        ]
        self.assertEqual(actual, api)

    def test_convert_df_to_api_with_name(self):
        definition = self.load_def("""{
              "objects": [
                {
                  "id": "MySchedule",
                  "name": "OVERRIDE-NAME",
                  "type": "Schedule",
                  "startDateTime": "2013-08-18T00:00:00",
                  "endDateTime": "2013-08-19T00:00:00",
                  "period": "1 day"
                }
            ]}""")
        actual = translator.definition_to_api_objects(definition)
        api = [
            {
                "name": "OVERRIDE-NAME",
                "id": "MySchedule",
                "fields": [
                    {
                        "key": "endDateTime",
                        "stringValue": "2013-08-19T00:00:00",
                    },
                    {"key": "period", "stringValue": "1 day"},
                    {
                        "key": "startDateTime",
                        "stringValue": "2013-08-18T00:00:00",
                    },
                    {"key": "type", "stringValue": "Schedule"},
                ],
            }
        ]
        self.assertEqual(actual, api)

    def test_objects_key_is_missing_raise_error(self):
        definition = self.load_def("""{"not-objects": []}""")
        with self.assertRaises(translator.PipelineDefinitionError):
            translator.definition_to_api_objects(definition)

    def test_missing_id_field(self):
        # Note that the 'id' key is missing.
        definition = self.load_def("""{
              "objects": [
                {
                  "name": "OVERRIDE-NAME",
                  "type": "Schedule",
                  "startDateTime": "2013-08-18T00:00:00",
                  "endDateTime": "2013-08-19T00:00:00",
                  "period": "1 day"
                }
            ]}""")
        with self.assertRaises(translator.PipelineDefinitionError):
            translator.definition_to_api_objects(definition)

    def test_list_value_with_strings(self):
        definition = self.load_def("""{"objects": [
            {
              "id" : "emrActivity",
              "type" : "EmrActivity",
              "name" : "Foo",
              "step" : ["s3://foo1", "s3://foo2", "s3://foo3"]
            }
        ]}""")
        actual = translator.definition_to_api_objects(definition)
        api = [
            {
                "name": "Foo",
                "id": "emrActivity",
                "fields": [
                    {"key": "step", "stringValue": "s3://foo1"},
                    {"key": "step", "stringValue": "s3://foo2"},
                    {"key": "step", "stringValue": "s3://foo3"},
                    {"key": "type", "stringValue": "EmrActivity"},
                ],
            }
        ]
        self.assertEqual(actual, api)

    def test_value_with_refs(self):
        definition = self.load_def("""{"objects": [
            {
              "id" : "emrActivity",
              "type" : "EmrActivity",
              "name" : "Foo",
              "step" : ["s3://foo1", {"ref": "otherValue"}, "s3://foo3"]
            }
        ]}""")
        actual = translator.definition_to_api_objects(definition)
        api = [
            {
                "name": "Foo",
                "id": "emrActivity",
                "fields": [
                    {"key": "step", "stringValue": "s3://foo1"},
                    {"key": "step", "refValue": "otherValue"},
                    {"key": "step", "stringValue": "s3://foo3"},
                    {"key": "type", "stringValue": "EmrActivity"},
                ],
            }
        ]
        self.assertEqual(actual, api)

    # These tests check the API -> DF conversion.
    def test_api_to_df(self):
        api = {
            "pipelineObjects": [
                {
                    "name": "S3ToS3Copy",
                    "id": "S3ToS3Copy",
                    "fields": [
                        {"key": "type", "stringValue": "CopyActivity"},
                        {"key": "schedule", "refValue": "CopyPeriod"},
                        {"key": "input", "refValue": "InputData"},
                        {"key": "output", "refValue": "OutputData"},
                    ],
                }
            ]
        }
        definition = translator.api_to_definition(api)
        self.assertEqual(
            definition,
            {
                'objects': [
                    {
                        'id': 'S3ToS3Copy',
                        'name': 'S3ToS3Copy',
                        'type': 'CopyActivity',
                        'schedule': {'ref': 'CopyPeriod'},
                        'input': {'ref': 'InputData'},
                        'output': {'ref': 'OutputData'},
                    }
                ]
            },
        )

    def test_api_to_df_with_dupe_keys(self):
        # Duplicate keys should be aggregated into a list.
        api = {
            "pipelineObjects": [
                {
                    "name": "S3ToS3Copy",
                    "id": "S3ToS3Copy",
                    "fields": [
                        {"key": "type", "stringValue": "CopyActivity"},
                        {"key": "schedule", "refValue": "CopyPeriod"},
                        {"key": "script", "stringValue": "value1"},
                        {"key": "script", "stringValue": "value2"},
                    ],
                }
            ]
        }
        definition = translator.api_to_definition(api)
        self.assertEqual(
            definition,
            {
                'objects': [
                    {
                        'id': 'S3ToS3Copy',
                        'name': 'S3ToS3Copy',
                        'type': 'CopyActivity',
                        'schedule': {'ref': 'CopyPeriod'},
                        'script': ['value1', 'value2'],
                    }
                ]
            },
        )

    def test_df_to_api_with_parameter_objects_missing(self):
        definition = self.load_def("""{"objects": [
            {
              "id" : "S3ToS3Copy",
              "type" : "CopyActivity",
              "schedule" : { "ref" : "CopyPeriod" },
              "input" : { "ref" : "InputData" },
              "output" : { "ref" : "OutputData" }
            }
            ]}""")
        try:
            parameters = translator.definition_to_api_parameters(definition)
            error = "Absence of parameter objects should not display anything"
            self.assertIsNone(parameters, error)
        except PipelineDefinitionError:
            self.fail("No exception should be thrown in absence of objects")

    def test_df_to_api_with_parameter_values_missing(self):
        definition = self.load_def("""{"objects": [
            {
              "id" : "S3ToS3Copy",
              "type" : "CopyActivity",
              "schedule" : { "ref" : "CopyPeriod" },
              "input" : { "ref" : "InputData" },
              "output" : { "ref" : "OutputData" }
            }
            ]}""")
        try:
            values = translator.definition_to_parameter_values(definition)
            error = "Absence of parameter values should not display anything"
            self.assertIsNone(values, error)
        except PipelineDefinitionError:
            self.fail("No exception should be thrown in absence of values")

    def test_df_to_api_with_parameter_objects(self):
        definition = self.load_def("""{"parameters": [
            {
              "id":"myOccurrenceNumber",
              "type":"Integer",
              "description":"Number of pipeline runs",
              "max":"24"
            }
            ]}""")
        actual = translator.definition_to_api_parameters(definition)
        api = [
            {
                "id": "myOccurrenceNumber",
                "attributes": [
                    {
                        "key": "description",
                        "stringValue": "Number of pipeline runs",
                    },
                    {"key": "max", "stringValue": "24"},
                    {"key": "type", "stringValue": "Integer"},
                ],
            }
        ]
        self.assertEqual(actual, api)

    def test_df_to_api_with_parameter_values(self):
        definition = self.load_def("""{"values": {
              "myMaximumRetries":"2",
              "myOccurrenceNumber":"3"
            }
            }""")
        actual = translator.definition_to_parameter_values(definition)
        api = [
            {"id": "myMaximumRetries", "stringValue": "2"},
            {"id": "myOccurrenceNumber", "stringValue": "3"},
        ]
        self.assertEqual(actual, api)

    def test_df_to_api_with_parameter_objects_array(self):
        definition = self.load_def("""{"parameters": [
            {
              "id":"myOccurrenceNumber",
              "type":"Integer",
              "description":["part1","part2"],
              "max":"24"
            }
            ]}""")
        actual = translator.definition_to_api_parameters(definition)
        api = [
            {
                "id": "myOccurrenceNumber",
                "attributes": [
                    {"key": "description", "stringValue": "part1"},
                    {"key": "description", "stringValue": "part2"},
                    {"key": "max", "stringValue": "24"},
                    {"key": "type", "stringValue": "Integer"},
                ],
            }
        ]
        self.assertEqual(actual, api)

    def test_df_to_api_with_parameter_values_array(self):
        definition = self.load_def("""{"values": {
              "myMaximumRetries":["value1", "value2"],
              "myOccurrenceNumber":"3"
            }
            }""")
        actual = translator.definition_to_parameter_values(definition)
        api = [
            {"id": "myMaximumRetries", "stringValue": "value1"},
            {"id": "myMaximumRetries", "stringValue": "value2"},
            {"id": "myOccurrenceNumber", "stringValue": "3"},
        ]
        self.assertEqual(actual, api)

    def test_api_to_df_with_parameter_objects_and_values(self):
        # Tests for proper handling of parameter value arrays as well
        api = {
            "parameterObjects": [
                {
                    "attributes": [
                        {"key": "max", "stringValue": "24"},
                        {
                            "key": "description",
                            "stringValue": "Number of pipeline runs",
                        },
                        {"key": "type", "stringValue": "Integer"},
                    ],
                    "id": "myOccurrenceNumber",
                },
                {
                    "attributes": [
                        {"key": "max", "stringValue": "3"},
                        {
                            "key": "description",
                            "stringValue": "number of retires",
                        },
                        {"key": "type", "stringValue": "Integer"},
                    ],
                    "id": "myMaximumRetries",
                },
            ],
            "parameterValues": [
                {"id": "myOccurrenceNumber", "stringValue": "4"},
                {"id": "myMaximumRetries", "stringValue": "2"},
                {"id": "myMaximumRetries", "stringValue": "3"},
            ],
            "pipelineObjects": [
                {
                    "fields": [
                        {
                            "key": "startDateTime",
                            "stringValue": "2011-01-01T00:00:00",
                        },
                        {
                            "key": "occurrences",
                            "stringValue": "#{myOccurrenceNumber}",
                        },
                        {"key": "period", "stringValue": "1 YEAR"},
                        {"key": "type", "stringValue": "Schedule"},
                    ],
                    "id": "CopyPeriod",
                    "name": "CopyPeriod",
                },
                {
                    "fields": [
                        {"key": "schedule", "refValue": "CopyPeriod"},
                        {"key": "retryDelay", "stringValue": "1 MINUTE"},
                        {"key": "command", "stringValue": "false"},
                        {
                            "key": "maximumRetries",
                            "stringValue": "#{myMaximumRetries}",
                        },
                        {"key": "workerGroup", "stringValue": "$PIPELINE_ID"},
                        {"key": "type", "stringValue": "ShellCommandActivity"},
                    ],
                    "id": "ShellCommand",
                    "name": "ShellCommand",
                },
            ],
        }
        expected = {
            "parameters": [
                {
                    "max": "24",
                    "type": "Integer",
                    "id": "myOccurrenceNumber",
                    "description": "Number of pipeline runs",
                },
                {
                    "max": "3",
                    "type": "Integer",
                    "id": "myMaximumRetries",
                    "description": "number of retires",
                },
            ],
            "objects": [
                {
                    "name": "CopyPeriod",
                    "period": "1 YEAR",
                    "occurrences": "#{myOccurrenceNumber}",
                    "startDateTime": "2011-01-01T00:00:00",
                    "type": "Schedule",
                    "id": "CopyPeriod",
                },
                {
                    "name": "ShellCommand",
                    "schedule": {"ref": "CopyPeriod"},
                    "retryDelay": "1 MINUTE",
                    "workerGroup": "$PIPELINE_ID",
                    "command": "false",
                    "type": "ShellCommandActivity",
                    "id": "ShellCommand",
                    "maximumRetries": "#{myMaximumRetries}",
                },
            ],
            "values": {
                "myMaximumRetries": ["2", "3"],
                "myOccurrenceNumber": "4",
            },
        }
        actual = translator.api_to_definition(api)
        self.assertEqual(expected, actual)



import pytest

from tests.backends.build_system.integration import VEnvWorkspace


@pytest.fixture
def workspace(tmp_path) -> VEnvWorkspace:
    return VEnvWorkspace(tmp_path)


@pytest.fixture(scope='module')
def exe_deps(tmpdir_factory):
    workspace = VEnvWorkspace(tmpdir_factory.mktemp('build_exe'))
    workspace.call_build_system("portable-exe", download_deps=True)
    return workspace


@pytest.fixture(scope='module')
def exe_no_deps(tmpdir_factory):
    workspace = VEnvWorkspace(tmpdir_factory.mktemp('build_exe'))
    workspace.install_dependencies()
    workspace.install_pyinstaller()
    workspace.call_build_system("portable-exe", download_deps=False)
    return workspace


@pytest.fixture(scope='module')
def sdist_deps(tmpdir_factory):
    workspace = VEnvWorkspace(tmpdir_factory.mktemp('system-sandbox'))
    workspace.call_build_system("system-sandbox", download_deps=True)
    return workspace


@pytest.fixture(scope='module')
def sdist_no_deps(tmpdir_factory):
    workspace = VEnvWorkspace(tmpdir_factory.mktemp('system-sandbox'))
    workspace.install_dependencies()
    workspace.call_build_system("system-sandbox", download_deps=False)
    return workspace
